import logging
import os
import pathlib
from multiprocessing import cpu_count
from typing import Union, Sequence, List

from pyproj import CRS

from ocsmesh.raster import Raster
from ocsmesh.hfun.hfun import Hfun
from ocsmesh.mesh.mesh import Mesh
from ocsmesh import utils

_logger = logging.getLogger(__name__)


class HfunCombine:

    def __init__(
            self,
            dem_files: Sequence[Union[str, os.PathLike]],
            out_file: Union[str, os.PathLike],
            out_format: str = "shapefile",
            mesh_file: Union[str, os.PathLike, None] = None,
            hmin: Union[float, None] = None,
            hmax: Union[float, None] = None,
            contours: List[List[float]] = None,
            constants: List[List[float]] = None,
            chunk_size: Union[int, None] = None,
            overlap: Union[int, None] = None,
            method: str = 'exact',
            nprocs: int = -1):


        self._base_exterior = None

        self._operation_info = {
                'dem_files': dem_files,
                'out_file': out_file,
                'out_format': out_format,
                'mesh_file': mesh_file,
                'hmin': hmin,
                'hmax': hmax,
                'contours': contours,
                'constants': constants,
                'chunk_size': chunk_size,
                'overlap': overlap,
                'method': method,
                'nprocs': nprocs
        }

    def run(self):

        dem_files = self._operation_info['dem_files']
        out_file = self._operation_info['out_file']
        out_format = self._operation_info['out_format']
        mesh_file = self._operation_info['mesh_file']
        hmin = self._operation_info['hmin']
        hmax = self._operation_info['hmax']
        contours = self._operation_info['contours']
        constants = self._operation_info['constants']
        chunk_size = self._operation_info['chunk_size']
        overlap = self._operation_info['overlap']
        method = self._operation_info['method']
        nprocs = self._operation_info['nprocs']

        nprocs = cpu_count() if nprocs == -1 else nprocs

        out_dir = pathlib.Path(out_file).parent
        out_dir.mkdir(exist_ok=True, parents=True)

        logging.info("Loading base mesh...")
        base_mesh = None
        if mesh_file:
            base_mesh = Mesh.open(mesh_file, crs="EPSG:4326")

        rast_list = []
        for dem_path in dem_files:

            logging.info(f"Loading raster {dem_path}...")
            rast = Raster(
                    dem_path, chunk_size=chunk_size, overlap=overlap)
            rast_list.append(rast)

        # Create Hfun
        logging.info("Creating Hfun from rasters...")
        hfun_collector = Hfun(
                rast_list, base_mesh=base_mesh,
                hmin=hmin, hmax=hmax, nprocs=nprocs, method=method)

        for contour in contours:
            logging.info("Adding contour refinement...")
            if len(contour) > 3:
                raise ValueError(
                    "Invalid format for contour specification."
                    " It should be level [expansion target-size].")

            level, expansion_rate, target_size = [
                    *contour, *[None]*(3-len(contour))]

            if level is None:
                raise ValueError(
                    "Invalid format for contour specification."
                    " It should be level [expansion target-size].")
            if expansion_rate is None:
                expansion_rate = 0.1
            if target_size is None:
                target_size = hmin

            hfun_collector.add_contour(
                level, expansion_rate, target_size)

        for lower_bound, target_size in constants:
            hfun_collector.add_constant_value(
                    value=target_size, lower_bound=lower_bound)

        self._write_to_file(
                out_format, out_file, hfun_collector, 'EPSG:4326')


    def _write_to_file(
            self, out_format, out_file, hfun_collector, crs):

        _logger.info(f"Writing for file ({out_format}) ...")

        # NOTE: Combined mesh from collector is always in EPSG:4326
        meshdata_hfun = hfun_collector.meshdata()
        dst_crs = CRS.from_user_input(crs)
        if meshdata_hfun.crs != dst_crs:
            _logger.info(f"Reprojecting hfun to ({crs}) ...")
            utils.reproject(meshdata_hfun, dst_crs)

        # TODO: Check for correct extension on out_file
        if out_format in ['2dm', 'sms']:
            # TODO: How to specify crs in 2dm file?
            mesh = Mesh(meshdata_hfun)
            mesh.write(out_file, format='2dm')

        else:
            raise NotImplementedError(f"Output type {out_format} is not supported")

        _logger.info("Done")



import sys
import unittest
from copy import deepcopy
from pathlib import Path
import shutil
import tempfile
import warnings

import geopandas as gpd
import numpy as np
import rasterio as rio
from shapely import geometry

import ocsmesh

from tests.api.test_common import (
    topo_2rast_1mesh,
)


class GeomType(unittest.TestCase):
    def setUp(self):
        self.tdir = Path(tempfile.mkdtemp())

        self.rast = self.tdir / 'rast_1.tif'
        self.mesh = self.tdir / 'mesh_1.gr3'

        rast_xy = np.mgrid[-1:0.1:0.1, -0.7:0.1:0.1]
        rast_xy = np.mgrid[0:1.1:0.1, -0.7:0.1:0.1]
        rast_z = np.ones_like(rast_xy[0]) * 10

        ocsmesh.utils.raster_from_numpy(
            self.rast, rast_z, rast_xy, 4326
        )

        meshdata = ocsmesh.utils.create_rectangle_mesh(
            nx=17, ny=7, holes=[40, 41], x_extent=(-1, 1), y_extent=(0, 1))
        meshdata.crs = 4326

        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', category=UserWarning, message='Input mesh has no CRS information'
            )
            mesh = ocsmesh.Mesh(meshdata)
            mesh.write(str(self.mesh), format='grd', overwrite=False)


    def tearDown(self):
        shutil.rmtree(self.tdir)


    def test_create_raster_geom(self):
        geom = ocsmesh.Geom(
            ocsmesh.Raster(self.rast),
            zmin=-100,
            zmax=10
        )
        self.assertTrue(isinstance(geom, ocsmesh.geom.raster.RasterGeom))


    def test_create_mesh_geom(self):
        geom = ocsmesh.Geom(
            ocsmesh.Mesh.open(self.mesh, crs=4326),
        )
        self.assertTrue(isinstance(geom, ocsmesh.geom.mesh.MeshGeom))


    def test_create_shape_geom(self):
        geom = ocsmesh.Geom(
            geometry.box(0, 0, 1, 1),
            crs=4326
        )
        self.assertTrue(isinstance(geom, ocsmesh.geom.shapely.PolygonGeom))


    def test_create_shape_geom_no_crs(self):
        self.assertRaises(
            TypeError,
            ocsmesh.Geom,
            geometry.box(0, 0, 1, 1),
        )


    def test_create_collector_geom(self):
        geom = ocsmesh.Geom(
            [self.rast],
            zmin=-100,
            zmax=10
        )
        self.assertTrue(isinstance(geom, ocsmesh.geom.collector.GeomCollector))


class GeomRaster(unittest.TestCase):
    def setUp(self):
        self.tdir = Path(tempfile.mkdtemp())
        self.rast1 = self.tdir / 'rast_1.tif'
        self.rast2 = self.tdir / 'rast_2.tif'
        self.mesh1 = self.tdir / 'mesh_1.gr3'
        topo_2rast_1mesh(self.rast1, self.rast2, self.mesh1)
        

    def tearDown(self):
        shutil.rmtree(self.tdir)


    def test_creation(self):

        rast1 = ocsmesh.Raster(self.rast1)
        geom_rast = ocsmesh.Geom(
            rast1,
            zmin=-100,
            zmax=10
        )

        geom_poly = geom_rast.get_multipolygon()
        self.assertTrue(
            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
        )

        geom_gs = geom_rast.geoseries()
        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))



class GeomCollector(unittest.TestCase):
    # NOTE: Testing a mixed collector geom indirectly tests almost
    # all the other types as it is currently calling all the underlying
    # geoms to calculate MeshData or polygons

    def setUp(self):
        self.tdir = Path(tempfile.mkdtemp())
        self.rast1 = self.tdir / 'rast_1.tif'
        self.rast2 = self.tdir / 'rast_2.tif'
        self.mesh1 = self.tdir / 'mesh_1.gr3'
        self.bx = geometry.box(-1, -2, 1, -1)
        topo_2rast_1mesh(self.rast1, self.rast2, self.mesh1)


    def tearDown(self):
        shutil.rmtree(self.tdir)


    def test_multi_path_input(self):
        geom_coll = ocsmesh.Geom(
            [self.rast1, self.rast2, self.mesh1],
            zmin=-100,
            zmax=10
        )

        geom_poly = geom_coll.get_multipolygon()
        self.assertTrue(
            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
        )

        geom_gs = geom_coll.geoseries()
        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))


    def test_multi_str_input(self):
        geom_coll = ocsmesh.Geom(
            [str(i) for i in [self.rast1, self.rast2, self.mesh1]],
            zmin=-100,
            zmax=10
        )

        geom_poly = geom_coll.get_multipolygon()
        self.assertTrue(
            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
        )

        geom_gs = geom_coll.geoseries()
        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))


    def test_multi_raster_input(self):

        rast1 = ocsmesh.Raster(self.rast1)
        rast2 = ocsmesh.Raster(self.rast2)
        geom_coll = ocsmesh.Geom(
            [rast1, rast2],
            zmin=-100,
            zmax=10
        )

        geom_poly = geom_coll.get_multipolygon()
        self.assertTrue(
            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
        )

        geom_gs = geom_coll.geoseries()
        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))


    def test_multi_mix_input(self):
        rast1 = ocsmesh.Raster(self.rast1)
        mesh1 = ocsmesh.Mesh.open(self.mesh1, crs=4326)
        geom_coll = ocsmesh.Geom(
            # TODO: Note supported yet
            [rast1, self.rast2, mesh1],
#            [rast1, self.rast2, mesh1, self.bx],
            zmin=-100,
            zmax=10
        )

        geom_poly = geom_coll.get_multipolygon()
        self.assertTrue(
            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
        )

        geom_gs = geom_coll.geoseries()
        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))


    def test_mesh_input(self):
        mesh1 = ocsmesh.Mesh.open(self.mesh1, crs=4326)
        geom_coll = ocsmesh.Geom(
            [mesh1],
            zmin=-100,
            zmax=10
        )

        geom_poly = geom_coll.get_multipolygon()
        self.assertTrue(
            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
        )

        geom_gs = geom_coll.geoseries()
        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))


    # TODO: Note supported yet
#    def test_shape_input(self):
#        geom_coll = ocsmesh.Geom(
#            [self.bx],
#            zmin=-100,
#            zmax=10
#        )
#
#
#        geom_poly = geom_coll.get_multipolygon()
#        self.assertTrue(
#            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
#        )
#
#        geom_gs = geom_coll.geoseries()
#        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))

    @unittest.skipIf(sys.platform.startswith("win"),
                     "Flaky I/O race condition on Windows multiprocessing")
    def test_add_patch(self):
        rast1 = ocsmesh.Raster(self.rast1)
        mesh1 = ocsmesh.Mesh.open(self.mesh1, crs=4326)
        geom_coll = ocsmesh.Geom(
            # TODO: Note supported yet
            [rast1, self.rast2, mesh1],
#            [rast1, self.rast2, mesh1, self.bx],
            zmin=-100,
            zmax=10
        )

        geom_coll.add_patch(
            shape=geometry.box(-0.25, -0.5, 0.5, -0.25),
            level=50,
        )

        geom_poly = geom_coll.get_multipolygon()
        self.assertTrue(
            isinstance(geom_poly, (geometry.Polygon, geometry.MultiPolygon))
        )

        geom_gs = geom_coll.geoseries()
        self.assertTrue(isinstance(geom_gs, gpd.GeoSeries))


    def test_add_patch_shape_should_overlap_all_rasters(self):
        rast1 = ocsmesh.Raster(self.rast1)
        mesh1 = ocsmesh.Mesh.open(self.mesh1, crs=4326)
        geom_coll = ocsmesh.Geom(
            [rast1, self.rast2, mesh1],
            zmin=-100,
            zmax=10
        )

        geom_coll.add_patch(
            shape=geometry.box(-0.75, -0.5, 0, -0.2),
            level=50,
        )

        self.assertRaises(
            ValueError,
            geom_coll.get_multipolygon
        )




if __name__ == '__main__':
    unittest.main()



import unittest
import tempfile
import platform
import shutil
import gc
import os
from pathlib import Path
import numpy as np
import numpy.testing as npt

from ocsmesh import Hfun, Raster
from ocsmesh.hfun.raster import HfunRaster
from ocsmesh.utils import raster_from_numpy



IS_WINDOWS = platform.system() == 'Windows'


class TestHfunCollectorExecution(unittest.TestCase):
    """
    Test the advanced execution features of HfunCollector, including
    serial vs parallel modes and resource cleanup.
    """


    def setUp(self):
        """Create a temporary directory and sample DEM files for testing."""
        self.tdir = Path(tempfile.mkdtemp())
        self.dem1_path = self.tdir / 'dem1.tif'
        self.dem2_path = self.tdir / 'dem2.tif'

        # Create a simple raster with a linear slope from -10 to 10
        # This makes it easy to verify the effects of refinements.
        grid_x, grid_y = np.mgrid[0:1:100j, 0:1:100j]
        dem_data = (grid_x * 20) - 10  # Values from -10 to 10

        # Create two identical DEM files for multi-raster tests
        raster_from_numpy(self.dem1_path, dem_data, (grid_x, grid_y), 4326)
        raster_from_numpy(self.dem2_path, dem_data.copy(), (grid_x, grid_y), 4326)

        # Create a list of raster objects for HfunCollector
        self.raster_list = [Raster(self.dem1_path), Raster(self.dem2_path)]


    def tearDown(self):
        """Remove the temporary directory and all its contents."""
        shutil.rmtree(self.tdir)


    @unittest.skipIf(IS_WINDOWS, 'Pickle tests not guaranteed stable on Windows due to I/O issues')
    def test_hfunraster_initial_value_logic(self):
        """
        Verify that HfunRaster correctly initializes from a file path.
        This is critical for the worker functions.
        """
        # 1. Create a raster with a known, unique value
        initial_hfun_path = self.tdir / 'initial_hfun.tif'
        initial_data = np.full((10, 10), 777.0, dtype=np.float32)
        grid_x, grid_y = np.mgrid[0:1:10j, 0:1:10j]
        raster_from_numpy(initial_hfun_path, initial_data, (grid_x, grid_y), 4326)

        # 2. Create an HfunRaster using this file as the initial_value
        base_raster = Raster(self.dem1_path)
        hfun = HfunRaster(
            raster=base_raster,
            initial_value=initial_hfun_path
        )

        # 3. Read the values and assert they match the initial file, not the default.
        loaded_values = hfun.get_values()
        self.assertEqual(loaded_values.shape, initial_data.shape)
        npt.assert_allclose(loaded_values, initial_data)
        # Ensure it's not the default "blank" value
        self.assertNotEqual(loaded_values[0, 0], np.finfo(np.float32).max)
        print("\nSUCCESS: HfunRaster correctly initializes from `initial_value`.")


    @unittest.skipIf(IS_WINDOWS, 'Pickle tests not guaranteed stable on Windows due to I/O issues')
    def test_execution_mode_property(self):
        """Test the behavior of the `execution_mode` property."""
        hfun = Hfun(self.raster_list, nprocs=2)

        # 1. Test that the default mode is 'serial' (lazy initialization)
        self.assertEqual(hfun.execution_mode, 'serial')

        # 2. Test setting the mode to 'parallel'
        hfun.execution_mode = 'parallel'
        self.assertEqual(hfun.execution_mode, 'parallel')

        # 3. Test that setting an invalid mode raises a ValueError
        with self.assertRaises(ValueError):
            hfun.execution_mode = 'invalid_mode'

        # 4. Test that a warning is issued if setting parallel with nprocs=1
        hfun_single_core = Hfun(self.raster_list, nprocs=1)
        with self.assertWarns(UserWarning):
            hfun_single_core.execution_mode = 'parallel'

    
    @unittest.skipIf(IS_WINDOWS, 'Pickle tests not guaranteed stable on Windows due to I/O issues')
    def test_work_dir_cleanup(self):
        """Verify that the temporary _work_dir is deleted when the object is destroyed."""
        hfun = Hfun(self.raster_list, nprocs=2)
        # Get the path to the temporary directory
        work_dir_path = hfun._work_dir

        # Assert that the directory exists after creation
        self.assertTrue(os.path.exists(work_dir_path))

        # Explicitly delete the object to trigger __del__
        del hfun
        # Encourage the garbage collector to run
        gc.collect()

        # Assert that the directory no longer exists
        self.assertFalse(os.path.exists(work_dir_path))

    
    @unittest.skipIf(IS_WINDOWS, 'Pickle tests not guaranteed stable on Windows due to I/O issues')
    def test_serial_vs_parallel_equivalence(self):
        """
        The main test: ensure that running refinements in serial and parallel
        modes produces numerically equivalent results.
        """
        nprocs = 2  # Use 2 cores for a simple parallel test

        # --- SERIAL EXECUTION ---
        hfun_serial = Hfun(self.raster_list, nprocs=nprocs, hmin=10, hmax=1000)
        # The default execution_mode is 'serial'
        
        # Add a chain of refinements
        hfun_serial.add_subtidal_flow_limiter(hmin=50, lower_bound=-5, upper_bound=5)
        hfun_serial.add_constant_value(value=200, lower_bound=5, upper_bound=10)
        
        print("\nRunning serial execution...")
        meshdata_serial = hfun_serial.meshdata()
        values_serial = meshdata_serial.values
        print("Serial execution finished.")

        # --- PARALLEL EXECUTION ---
        hfun_parallel = Hfun(self.raster_list, nprocs=nprocs, hmin=10, hmax=1000)
        hfun_parallel.execution_mode = 'parallel'  # Explicitly set to parallel
        
        # Add the exact same chain of refinements
        hfun_parallel.add_subtidal_flow_limiter(hmin=50, lower_bound=-5, upper_bound=5)
        hfun_parallel.add_constant_value(value=200, lower_bound=5, upper_bound=10)

        print("Running parallel execution...")
        meshdata_parallel = hfun_parallel.meshdata()
        values_parallel = meshdata_parallel.values
        print("Parallel execution finished.")
        
        # --- COMPARISON ---
        # NOTE: Due to minor floating point differences in meshing algorithms,
        # the vertex count and values might not be bit-for-bit identical.
        # The most robust comparison is to check key statistical properties.

        # 1. Check if the number of mesh nodes is very similar
        self.assertAlmostEqual(len(values_serial), len(values_parallel), delta=len(values_serial) * 0.01)

        # 2. Check if the min, max, and mean of the size function are almost identical
        npt.assert_allclose(np.min(values_serial), np.min(values_parallel), rtol=1e-5)
        npt.assert_allclose(np.max(values_serial), np.max(values_parallel), rtol=1e-5)
        npt.assert_allclose(np.mean(values_serial), np.mean(values_parallel), rtol=1e-5)



if __name__ == '__main__':
    unittest.main()



"""This module defines a base type for mesh objects.
"""
import warnings

import numpy as np
import numpy.typing as npt

from ocsmesh.internal import MeshData

class BaseMesh:
    """Base class for mesh types in OCSMesh.

    Attributes
    ----------
    meshdata : MeshData
        Refernece to underlying mesh object.
    coord : array-like
        Coordinates of mesh nodes.

    Methods
    -------
    """

    @property
    def msh_t(self) -> MeshData:
        warnings.warn("Use meshdata(...) instead!", DeprecationWarning)
        return self.meshdata


    @property
    def meshdata(self) -> MeshData:
        """Read-only property returning reference to the mesh data

        Notes
        -----
        The property is read-only, however the returned value is not.
        That means if the user mutates the returned MeshData
        object, it will affect the mesh.
        """

        #pylint: disable=E1101
        return self._meshdata

    @property
    def coord(self) -> npt.NDArray[np.float32]:
        """Read-only property for coordinates of mesh points

        Raises
        ------
        If the number of mesh dimensions is not 2 or 3.

        Notes
        -----
        The property is read-only, however the returned value is not.
        That means if the user mutates the returned numpy array
        object, it will affect the mesh.
        """

        return self.meshdata.coords



"""This module defines size function collector.
`
Size function collector objects accepts a list of valid basic `Hfun`
inputs and creates an object that merges the results of all the
other types of size functions, e.g. mesh-based and raster-based.

Notes
-----
This enables the user to define size on multiple DEM and mesh
without having to worry about the details of merging the output size
functions defined on each DEM or mesh.
"""
import os
import shutil
import gc
import logging
import warnings
import tempfile
from pathlib import Path
from time import time
from multiprocessing import Pool, cpu_count
from copy import copy, deepcopy
from typing import (
    Union, Sequence, List, Tuple, Iterable, Any, Optional, Callable)
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np
import numpy.typing as npt
import geopandas as gpd
from pyproj import CRS, Transformer
from shapely.geometry import (
    MultiLineString,
    LineString,
    MultiPolygon,
    Polygon,
    GeometryCollection,
)
from shapely import ops
from rasterio.transform import from_origin
from rasterio.warp import reproject, Resampling
import rasterio

from ocsmesh import utils
from ocsmesh.internal import MeshData
from ocsmesh.hfun.base import BaseHfun
from ocsmesh.hfun.raster import HfunRaster
from ocsmesh.hfun.mesh import HfunMesh
from ocsmesh.mesh.mesh import Mesh, EuclideanMesh2D
from ocsmesh.mesh.base import BaseMesh
from ocsmesh.raster import Raster, get_iter_windows
from ocsmesh.features.contour import Contour
from ocsmesh.features.patch import Patch
from ocsmesh.features.linefeature import LineFeature
from ocsmesh.features.channel import Channel
from ocsmesh.features.constraint import (
    TopoConstConstraint,
    TopoFuncConstraint,
    CourantNumConstraint,
    RegionConstraint,
)

CanCreateSingleHfun = Union[Raster, EuclideanMesh2D]
CanCreateMultipleHfun = Iterable[Union[CanCreateSingleHfun, str]]
CanCreateHfun = Union[CanCreateSingleHfun, CanCreateMultipleHfun]

SizeFuncList = Sequence[Union[HfunRaster, HfunMesh]]

RASTER_CONSTR = (
    TopoConstConstraint,
    TopoFuncConstraint,
    CourantNumConstraint,
)

_logger = logging.getLogger(__name__)

class _RefinementContourInfoCollector:
    """Collection for contour refinement specification

    Accumulates information about the specified contour refinements
    to be applied later when computing the return size function.
    Provides interator interface for looping over the collection items.
    """

    def __init__(self) -> None:
        self._contours_info = {}

    def add(self, contour_defn: Contour, **size_info: Any) -> None:
        """Add contour specification to the collection

        Parameters
        ----------
        contour_defn : Contour
            The level at which contour lines need to be extracted.
        size_info : dict
            Information related to contour application such as
            target size, rate, etc.

        Returns
        -------
        None
        """

        self._contours_info[contour_defn] = size_info

    def __iter__(self) -> Tuple[Contour, dict]:
        """Iterator method for this collection object

        Yields
        ------
        defn : Contour
            Contour definition added to the collection.
        info : dict
            Dictionary of contour refinement specifications.
        """

        for defn, info in self._contours_info.items():
            yield defn, info




class _RefinementContourCollector:
    """Collection for extracted refinement contours

    Extracts and stores on the disk the contours specified by the user.
    Provides interator interface for looping over the collection items.
    """

    def __init__(
            self,
            contours_info: _RefinementContourInfoCollector
            ) -> None:
        """Initialize the collection object with empty output list.

        Parameters
        ----------
        contours_info : _RefinementContourInfoCollector
            Handle to the collection of user specified contours
            specification.
        """

        self._contours_info = contours_info
        self._container: List[Union[Tuple, None]] = []

    def calculate(
            self,
            source_list: Iterable[HfunRaster],
            out_path: Union[Path, str]
            ) -> None:
        """Extract specified contours and store on disk in `out_path`.

        Parameters
        ----------
        source_list : list of HfunRaster
            List of raster size functions from which the contours
            must be calculated.
        out_path : path-like
            Path for storing calculated contours and their crs data.

        Returns
        -------
        None
        """

        out_dir = Path(out_path)
        out_dir.mkdir(exist_ok=True, parents=True)
        file_counter = 0
        pid = os.getpid()
        self._container.clear()
        for contour_defn, size_info in self._contours_info:
            if not contour_defn.has_source:
                # Copy so that in case of a 2nd run the no-source
                # contour still gets all current sources
                contour_defn = copy(contour_defn)
                for source in source_list:
                    contour_defn.add_source(source)

            for contour, crs in contour_defn.iter_contours():
                file_counter = file_counter + 1
                feather_path = out_dir / f"contour_{pid}_{file_counter}.feather"
                crs_path = out_dir / f"crs_{pid}_{file_counter}.json"
                gpd.GeoDataFrame(
                    { 'geometry': [contour],
                      'expansion_rate': size_info['expansion_rate'],
                      'target_size': size_info['target_size'],
                    },
                    crs=crs).to_feather(feather_path)
                gc.collect()
                with open(crs_path, 'w') as fp:
                    fp.write(crs.to_json())
                self._container.append((feather_path, crs_path))


    def __iter__(self) -> gpd.GeoDataFrame:
        """Iterator method for this collection object

        Yields
        ------
        gpd.GeoDataFrame
            Data from containing the extracted contour line and the
            CRS information.
        """

        for raster_data in self._container:
            feather_path, crs_path = raster_data
            gdf = gpd.read_feather(feather_path)
            with open(crs_path) as fp:
                gdf.set_crs(CRS.from_json(fp.read()))
            yield gdf




class _ConstantValueContourInfoCollector:
    """Collection for constant value refinement specification"""

    def __init__(self) -> None:
        self._contours_info = {}

    def add(self,
            src_idx: Optional[Sequence[int]],
            contour_defn0: Contour,
            contour_defn1: Contour,
            value: float
            ) -> None:
        """Add a new fixed-value refinement specification to the spec.

        Parameters
        ----------
        src_idx : tuple of int or None
            Indices of sources (indexed based on all `HfunCollector`
            **not** just rasters) on which constant value refinement
            must be applied.
        contour_defn0 : Contour
            Lower bound of region to apply constant value refinement.
        contour_defn1 : Contour
            Upper bound of region to apply constant value refinement.
        value : float
            Fixed-value to be applied as refinement.

        Returns
        -------
        None
        """

        srcs = tuple(src_idx) if src_idx is not None else None
        self._contours_info[
                (srcs, contour_defn0, contour_defn1)] = value

    def __iter__(self) -> Tuple[Tuple[Sequence[int], Contour, Contour], dict]:
        """Iterator method for this collection object

        Yields
        ------
        defn : Tuple[Sequence[int], Contour, Contour]
            The lower and upper bound contours definitions provided
            by the user as well as the list of source.
        info : dict
            Dictionary of specifications for constant value refinement.
        """

        for defn, info in self._contours_info.items():
            yield defn, info



class _RefinementShapeInfoCollector:
    """Collection for shape (patch or line) refinement specifications"""

    def __init__(self) -> None:
        self._shape_info = {}

    def add(self,
            shape_defn: Union[Patch, LineFeature],
            **size_info: Any
        ) -> None:
        """Add shape refinement specifications to the collection

        Parameters
        ----------
        shape_defn : Patch or LineFeature
            Shape of the region to apply the refinement during
            application.
        size_info : dict
            Information related to shape application such as
            target size, rate, etc.

        Returns
        -------
        None
        """

        self._shape_info[shape_defn] = size_info

    def __iter__(self) -> Tuple[Union[Patch, LineFeature], dict]:
        """Iterator method for this collection object

        Yields
        ------
        defn : Patch or LineFeature
            Object representing the shape of refinement
        info : dict
            Dictionary of specifications for shape refinement.
        """

        for defn, info in self._shape_info.items():
            yield defn, info



class _FlowLimiterInfoCollector:
    """Collection for subtidal flow limiter refinement spec."""

    def __init__(self) -> None:
        self._flow_lim_info = []

    def add(self,
            src_idx: Optional[Sequence[int]],
            hmin: float,
            hmax: float,
            upper_bound: float,
            lower_bound: float
            ) -> None:
        """Add subtidal flow limiter refinement spec to the collection.

        Parameters
        ----------
        src_idx : tuple of int or None
            Indices of sources (indexed based on all `HfunCollector`
            **not** just rasters) on which subtidal flow limiter
            refinement must be applied.
        hmin : float
            Minimum mesh size to be applied based on limiter
            calculations.
        hmax : float
            Maximum mesh size to be applied based on limiter
            calculations.
        upper_bound : float
            Elevation upper bound of the area that the limiter
            refinement is applied.
        lower_bound : float
            Elevation lower bound of the area that the limiter
            refinement is applied.

        Returns
        -------
        None
        """

        srcs = tuple(src_idx) if src_idx is not None else None
        self._flow_lim_info.append(
                (src_idx, hmin, hmax, upper_bound, lower_bound))

    def __iter__(self) -> Tuple[Sequence[int], float, float, float, float]:
        """Iterator method for this collection object

        Yields
        ------
        src_idx : tuple of int
            Similar to `add`
        hmin : float
            Similar to `add`
        hmax : float
            Similar to `add`
        ub : float
            Similar to `add`
        lb : float
            Similar to `add`
        """

        for src_idx, hmin, hmax, ub, lb in self._flow_lim_info:
            yield src_idx, hmin, hmax, ub, lb



class _ChannelRefineInfoCollector:
    """Collection for channel refinement specifications"""

    def __init__(self) -> None:
        self._ch_info_dict = {}

    def add(self,
            channel_defn: Channel,
            **size_info: Any
            ) -> None:
        """Add channel refinement spec to the collection.

        Parameters
        ----------
        channel_defn : Channel
            Definition of channel detection specification.
        size_info : dict
            Information related to channel refinement application
            such as target size, rate, etc.

        Returns
        -------
        None
        """

        self._ch_info_dict[channel_defn] = size_info

    def __iter__(self) -> Tuple[Channel, dict]:
        """Iterator method for this collection object

        Yields
        ------
        defn : Channel
            Similar to `channel_defn` in `add`
        info : dict
            Similar to `size_info` in `add`
        """

        for defn, info in self._ch_info_dict.items():
            yield defn, info


class _ChannelRefineCollector:
    """Collection for extracted refinement channels"""

    def __init__(self, channels_info) -> None:
        """Initialize the collection object with empty output list.

        Parameters
        ----------
        channels_info : _ChannelRefineInfoCollector
            Handle to the collection of user specified channel
            refinement specification.
        """

        self._channels_info = channels_info
        self._container: List[Union[Tuple, None]] = []

    def calculate(
            self,
            source_list,
            out_path
            ) -> None:
        """Extract specified channels and store on disk in `out_path`.

        Parameters
        ----------
        source_list : list of HfunRaster
            List of raster size functions from which the channels
            must be calculated.
        out_path : path-like
            Path for storing calculated channels and their crs data.

        Returns
        -------
        None
        """

        out_dir = Path(out_path)
        out_dir.mkdir(exist_ok=True, parents=True)
        file_counter = 0
        pid = os.getpid()
        self._container.clear()
        for channel_defn, size_info in self._channels_info:
            if not channel_defn.has_source:
                # Copy so that in case of a 2nd run the no-source
                # channel still gets all current sources
                channel_defn = copy(channel_defn)
                for source in source_list:
                    channel_defn.add_source(source)

            for channels, crs in channel_defn.iter_channels():
                file_counter = file_counter + 1
                feather_path = out_dir / f"channels_{pid}_{file_counter}.feather"
                crs_path = out_dir / f"crs_{pid}_{file_counter}.json"
                gpd.GeoDataFrame(
                    { 'geometry': [channels],
                      'expansion_rate': size_info['expansion_rate'],
                      'target_size': size_info['target_size'],
                    },
                    crs=crs).to_feather(feather_path)
                gc.collect()
                with open(crs_path, 'w') as fp:
                    fp.write(crs.to_json())
                self._container.append((feather_path, crs_path))

    def __iter__(self):
        """Iterator method for this collection object

        Yields
        ------
        gpd.GeoDataFrame
            Data from containing the extracted channel polygons and
            the CRS information.
        """

        for raster_data in self._container:
            feather_path, crs_path = raster_data
            gdf = gpd.read_feather(feather_path)
            with open(crs_path) as fp:
                gdf.set_crs(CRS.from_json(fp.read()))
            yield gdf


class _ConstraintInfoCollector:
    """Collection for the applied constraints"""

    def __init__(self) -> None:
        self._constraints_info = []

    def add(self,
            src_idx: Optional[Sequence[int]],
            constraint: Union[TopoConstConstraint, TopoFuncConstraint]
            ) -> None:
        """Add size function constraint spec to the collection.

        Parameters
        ----------
        src_idx : tuple of int or None
            Indices of sources (indexed based on all `HfunCollector`
            **not** just rasters) on which constant value refinement
            must be applied.
        constraint : TopoConstConstraint or TopoFuncConstraint
            The constraint definition object.

        Returns
        -------
        None
        """

        srcs = tuple(src_idx) if src_idx is not None else None
        self._constraints_info.append((srcs, constraint))

    def __iter__(self) -> Union[TopoConstConstraint, TopoFuncConstraint]:
        """Iterator method for this collection object

        Yields
        ------
        TopoConstConstraint or TopoFuncConstraint
            The constraint object provided in `add`
        """

        yield from self._constraints_info


    def apply(self, hfun_list, per_hfun=True):
        for in_idx, hfun in enumerate(hfun_list):
            is_raster = isinstance(hfun, HfunRaster)
            constraint_list = []
            for src_idx, constraint_defn in self:
                if per_hfun and src_idx is not None and in_idx not in src_idx:
                    continue

                if isinstance(constraint_defn, RASTER_CONSTR) and not is_raster:
                    continue

                constraint_list.append(constraint_defn)

            if constraint_list:
                hfun.apply_constraints(constraint_list)


def _flow_limiter_task_worker(task: dict):
    """
    A self-contained worker that operates ONLY on file paths.
    """

    # 1. Unpack the simple, pickleable task description
    original_index = task['original_index']
    hfun_input_path = task['hfun_input_path']
    topo_input_path = task['topo_input_path']
    output_path = task['output_path']
    global_hmin = task['global_hmin']
    global_hmax = task['global_hmax']
    limiter_params_list = task['limiter_params']

    # Let's initialize directly from the input path.

    # 2. Create the necessary Raster and HfunRaster instances INSIDE the worker.
    topo_raster = Raster(topo_input_path)
    worker_hfun = HfunRaster(
        raster=topo_raster,
        hmin=global_hmin,
        hmax=global_hmax,
        verbosity=0,
        initial_value=hfun_input_path
    )

    # 3. Apply all the required flow limiter refinements.
    #    Each call will modify the worker_hfun's internal state (_tmpfile).
    for params in limiter_params_list:
        worker_hfun.add_subtidal_flow_limiter(
            hmin=params['hmin'],
            hmax=params['hmax'],
            lower_bound=params['zmin'],
            upper_bound=params['zmax']
        )

    # 4. Explicitly save the final state of the worker's
    #    object to the designated output path.
    worker_hfun.save(output_path)

    # 5. The work is done. Return the simple result dictionary.
    return {
        'status': 'success',
        'original_index': original_index,
        'output_path': output_path
    }


def _const_val_task_worker(task: dict):
    """
    A self-contained worker for applying constant value refinements.
    It reads the state from the previous step, applies its rules, and saves
    a new file with the result.
    """

    # 1. Unpack the simple, pickleable task description
    original_index = task['original_index']
    hfun_input_path = task['hfun_input_path']
    topo_input_path = task['topo_input_path']
    output_path = task['output_path']
    global_hmin = task['global_hmin']
    global_hmax = task['global_hmax']
    const_val_rules = task['const_val_rules']

    # 2. Create the necessary Raster and HfunRaster instances INSIDE the worker.
    #    This is the "Wrap Existing Painting" mode.
    topo_raster = Raster(topo_input_path)
    worker_hfun = HfunRaster(
        raster=topo_raster,
        hmin=global_hmin,
        hmax=global_hmax,
        verbosity=0,
        initial_value=hfun_input_path
    )

    # 3. Apply all the required constant value refinements for this raster.
    for rule in const_val_rules:
        worker_hfun.add_constant_value(
            value=rule['value'],
            lower_bound=rule['lower_bound'],
            upper_bound=rule['upper_bound']
        )

    # 4. Explicitly save the final state of the worker's object to the
    #    designated output path.
    worker_hfun.save(output_path)

    # 5. The work is done. Return a simple result dictionary.
    return {
        'status': 'success',
        'original_index': original_index,
        'output_path': output_path
    }


class HfunCollector(BaseHfun):
    """Define size function based on multiple inputs of different types

    Attributes
    ----------

    Methods
    -------
    meshdata()
        Return mesh sizes interpolated on an size-optimized
        unstructured mesh
    add_topo_bound_constraint(...)
        Add size fixed-per-point value constraint to the area
        bounded by specified bounds with expansion/contraction
        rate `rate` specified. This refinement is only applied on
        rasters with specified indices. The index is w.r.t the
        full input list for collector object creation.
    add_topo_func_constraint(upper_bound=np.inf, lower_bound=-np.inf,
                             value_type='min', rate=0.01)
        Add size value constraint based on function of depth/elevation
        to the area bounded by specified bounds with the expansion or
        contraction rate `rate` specified. This constraint is only
        applied on rasters with specified indices. The index is w.r.t
        the full input list for collector object creation.
    add_region_constraint(...)
        Add a constraint on size based on a specified region
    add_patch(...)
        Add a region of fixed size refinement with optional expansion
        rate for points outside the region to achieve smooth size
        transition.
    add_contour(...)
        Add refinement based on contour lines auto-extrcted from the
        underlying raster data. The size is calculated based on the
        specified `rate`, `target_size` and the distance from the
        extracted feature line. For refinement contours are extracted
        only from raster inputs, but are applied on all input size
        function bases.
    add_channel(...)
        Add refinement for auto-detected narrow domain regions.
        Optionally use an expansion rate for points outside detected
        narrow regions for smooth size transition.
    add_subtidal_flow_limiter(...)
        Add mesh size refinement based on the value as well as
        gradient of the topography within the region between
        specified by lower and upper bound on topography.
        This refinement is only applied on rasters with specified
        indices. The index is w.r.t the full input list for collector
        object creation.
    add_constant_value(...)
        Add fixed size mesh refinement in the region specified by
        upper and lower bounds on topography.  This refinement is
        only applied on rasters with specified indices. The index is
        w.r.t the full input list for collector object creation.

    Notes
    -----
    All the refinements and constraints of this collector size
    function are applied lazily. That means the size values are **not**
    evaluated at the time they are called. Instead the effect of
    all these refinements and constraints on the size is calculated
    when `meshdata()` method is called.

    Two distinct algorithms are implemented for storing the size
    function values during evaluation and before creating the
    "background mesh" on which sizes are specified. Currently
    the difference between algorithms are only due to how raster
    inputs to the collector are processed. The **exact** algorithm
    is more computationally expensive; it processes all refinements
    on the original rasters and applies the indivitually on all
    of those individual rasters; this results in exponential time
    calculation of contours or raster features as the extracted
    features on all rasters must be applied on all rasters. Application
    of features usually involve distance calculation using a tree
    and can be very expensive when many rasters-features are involved.
    The **fast** approach is less exact and can use more memory, but
    it is much faster. The approach it takes is to still extract
    the raster features individually but then apply it to a lower
    resolution large raster that covers all the input rasters.
    """

    def __init__(
            self,
            in_list: CanCreateMultipleHfun,
            base_mesh: Optional[Mesh] = None,
            hmin: Optional[float] = None,
            hmax: Optional[float] = None,
            nprocs: Optional[int] = None,
            verbosity: int = 0,
            method: Literal['exact', 'fast'] = 'exact',
            base_as_hfun: bool = True,
            base_shape: Optional[Union[Polygon, MultiPolygon]] = None,
            base_shape_crs: Union[str, CRS] = 'EPSG:4326'
            ) -> None:
        """Initialize a collector size function object

        Parameters
        ----------
        in_list : CanCreateMultipleHfun
        base_mesh : Mesh or None, default=None
        hmin : float or None, default=None
        hmax : float or None, default=None
        nprocs : int or None, default=None
        verbosity : int, default=0
        method : {'exact', 'fast'}, default='exact
        base_as_hfun : bool, default=True
        base_shape: Polygon or MultiPolygon or None, default=None
        base_shape_crs: str or CRS, default='EPSG:4326'
        """

        # NOTE: Input Hfuns and their Rasters can get modified

         # Add a persistent working directory for this instance's outputs
        self._work_dir = tempfile.mkdtemp(prefix='hfun_collector_')
        # Check nprocs
        nprocs = -1 if nprocs is None else nprocs
        nprocs = cpu_count() if nprocs == -1 else nprocs

        self._applied = False
        self._size_info = {'hmin': hmin, 'hmax': hmax}
        self._nprocs = nprocs
        self._hfun_list = []
        self._method = method

        self._base_shape = base_shape
        self._base_shape_crs = CRS.from_user_input(base_shape_crs)
        self._base_as_hfun = base_as_hfun

        # NOTE: Base mesh has to have a crs otherwise HfunMesh throws
        # exception
        self._base_mesh = None
        if base_mesh:
            self._base_mesh = HfunMesh(base_mesh)
            if self._base_as_hfun:
                self._base_mesh.size_from_mesh()

        self._contour_info_coll = _RefinementContourInfoCollector()
        self._contour_coll = _RefinementContourCollector(
                self._contour_info_coll)

        self._const_val_contour_coll = _ConstantValueContourInfoCollector()

        self._refine_patch_info_coll = _RefinementShapeInfoCollector()
        self._refine_line_info_coll = _RefinementShapeInfoCollector()

        self._flow_lim_coll = _FlowLimiterInfoCollector()

        self._ch_info_coll = _ChannelRefineInfoCollector()
        self._channels_coll = _ChannelRefineCollector(
                self._ch_info_coll)

        self._constraint_info_coll = _ConstraintInfoCollector()

        self._type_chk(in_list)

        # TODO: Interpolate max size on base mesh basemesh?
        #
        # TODO: CRS considerations

        for in_item in in_list:
            # Add supports(ext) to each hfun type?

            if isinstance(in_item, BaseHfun):
                hfun = in_item
            # pylint: disable=R0801
            elif isinstance(in_item, Raster):
                if self._base_shape:
                    clip_shape = self._base_shape
                    if not self._base_shape_crs.equals(in_item.crs):
                        transformer = Transformer.from_crs(
                            self._base_shape_crs, in_item.crs, always_xy=True)
                        clip_shape = ops.transform(
                                transformer.transform, clip_shape)
                    try:
                        in_item.clip(clip_shape)
                    except ValueError as err:
                        # This raster does not intersect shape
                        _logger.debug(err)
                        continue

                elif self._base_mesh:
                    try:
                        in_item.clip(self._base_mesh.mesh.get_bbox(crs=in_item.crs))
                    except ValueError as err:
                        # This raster does not intersect shape
                        _logger.debug(err)
                        continue

                hfun = HfunRaster(in_item, **self._size_info)

            elif isinstance(in_item, EuclideanMesh2D):
                hfun = HfunMesh(in_item)

            elif isinstance(in_item, (str, Path)):
                in_item = str(in_item)
                if in_item.endswith('.tif'):
                    raster = Raster(in_item)
                    if self._base_shape:
                        clip_shape = self._base_shape
                        if not self._base_shape_crs.equals(raster.crs):
                            transformer = Transformer.from_crs(
                                self._base_shape_crs, raster.crs, always_xy=True)
                            clip_shape = ops.transform(
                                    transformer.transform, clip_shape)
                        try:
                            in_item.clip(clip_shape)
                        except ValueError as err:
                            # This raster does not intersect shape
                            _logger.debug(err)
                            continue

                    elif self._base_mesh:
                        try:
                            raster.clip(self._base_mesh.mesh.get_bbox(crs=raster.crs))
                        except ValueError as err:
                            # This raster does not intersect shape
                            _logger.debug(err)
                            continue

                    hfun = HfunRaster(raster, **self._size_info)

                elif in_item.endswith(
                        ('.14', '.grd', '.gr3', '.msh', '.2dm')):
                    mesh = Mesh.open(in_item)
                    hfun = HfunMesh(mesh)

                else:
                    raise TypeError("Input file extension not supported!")

            self._hfun_list.append(hfun) # pylint: disable=E0606


    def __del__(self):
        if hasattr(self, '_work_dir') and os.path.exists(self._work_dir):
            shutil.rmtree(self._work_dir, ignore_errors=True)


    def meshdata(self, **kwargs) -> MeshData:
        """Interpolates mesh size functions on an unstructred mesh

        Parameters
        ----------
        **kwargs : dict
            Arguments passed down to the underlying Hfun classes (e.g.
            mesh_engine='gmsh', stride=...).
        """

        # Just dummy object
        composite_hfun = MeshData([[0,0]])

        if self._method == 'exact':
            self._apply_features()

            with tempfile.TemporaryDirectory() as temp_dir:
                # Pass kwargs if needed
                hfun_path_list = self._write_hfun_to_disk(temp_dir, **kwargs)
                composite_hfun = self._get_hfun_composite(hfun_path_list)

        elif self._method == 'fast':

            with tempfile.TemporaryDirectory() as temp_dir:
                rast = self._create_big_raster(temp_dir)
                hfun = self._apply_features_fast(rast)
                # Pass kwargs (like mesh_engine) down to the composite generator
                composite_hfun = self._get_hfun_composite_fast(hfun, **kwargs)

                del rast
                del hfun

        else:
            raise ValueError(f"Invalid method specified: {self._method}")

        return composite_hfun


    def add_topo_bound_constraint(
            self,
            value: Union[float, npt.NDArray[np.float32]],
            upper_bound: float = np.inf,
            lower_bound: float = -np.inf,
            value_type: Literal['min', 'max'] = 'min',
            rate: float = 0.01,
            source_index: Union[List[int], int, None] = None
            ) -> None:
        """Add a fixed-value or fixed-matrix constraint.

        Add a fixed-value or fixed-matrix constraint to the region
        of the size function specified by lower and upper elevation
        of the underlying DEM. Optionally a `rate` can be specified
        to relax the constraint gradually outside the bounds.

        Parameters
        ----------
        value : float or array-like
            A single fixed value or array of values to be used for
            mesh size if condition is not met based on `value_type`.
            In case of an array the dimensions must match or be
            broadcastable to the raster grid.
        upper_bound : float, default=np.inf
            Maximum elevation to cut off the region where the
            constraint needs to be applied
        lower_bound : float, default=-np.inf
            Minimum elevation to cut off the region where the
            constraint needs to be applied
        value_type : {'min', 'max'}, default='min'
            Type of contraint. If 'min', it means the mesh size
            should not be smaller than the specified `value` at each
            point.
        rate : float, default=0.01
            Rate of relaxation of constraint outside the region defined
            by `lower_bound` and `upper_bound`.
        source_index : int or list of ints or None, default=None
            The index of entries from the input list argument
            of the constructor of collector size function. If `None`
            all input rasters are used.

        Returns
        -------
        None
        """

        self._applied = False

        constraint_defn = TopoConstConstraint(
            value, upper_bound, lower_bound, value_type, rate)

        if source_index is not None and not isinstance(source_index, (tuple, list)):
            source_index = [source_index]
        self._constraint_info_coll.add(source_index, constraint_defn)


    def add_topo_func_constraint(
            self,
            func: Callable[[npt.NDArray[np.float32]], npt.NDArray[np.float32]]
                = lambda i: i / 2.0,
            upper_bound: float = np.inf,
            lower_bound: float = -np.inf,
            value_type: Literal['min', 'max'] = 'min',
            rate: float = 0.01,
            source_index: Union[List[int], int, None] = None
            ) -> None:
        """Add constraint based on a function of the topography

        Add a constraint based on the provided function `func` of
        the topography and apply to the region of the size function
        specified by lower and upper elevation of the underlying DEM.
        Optionally a `rate` can be specified to relax the constraint
        gradually outside the bounds.

        Parameters
        ----------
        func : callable
            A function to be applied on the topography to acquire the
            values to be used for mesh size if condition is not met
            based on `value_type`.
        upper_bound : float, default=np.inf
            Maximum elevation to cut off the region where the
            constraint needs to be applied
        lower_bound : float, default=-np.inf
            Minimum elevation to cut off the region where the
            constraint needs to be applied
        value_type : {'min', 'max'}, default='min'
            Type of contraint. If 'min', it means the mesh size
            should not be smaller than the value calculated from the
            specified `func` at each point.
        rate : float, default=0.01
            Rate of relaxation of constraint outside the region defined
            by `lower_bound` and `upper_bound`.
        source_index : int or list of ints or None, default=None
            The index of entries from the input list argument
            of the constructor of collector size function. If `None`
            all input rasters are used.

        Returns
        -------
        None
        """

        self._applied = False

        constraint_defn = TopoFuncConstraint(
            func, upper_bound, lower_bound, value_type, rate)

        if source_index is not None and not isinstance(source_index, (tuple, list)):
            source_index = [source_index]
        self._constraint_info_coll.add(source_index, constraint_defn)


    def add_courant_num_constraint(
            self,
            upper_bound: float = 0.9,
            lower_bound: Optional[float] = None,
            timestep: float = 150,
            wave_amplitude: float = 2,
            source_index: Union[List[int], int, None] = None
            ) -> None:
        """Add constraint based on approximated Courant number bounds


        Parameters
        ----------
        upper_bound : float, default=0.9
            Maximum Courant number to allow on this mesh size function
        lower_bound : float or None, default=None
            Minimum Courant number to allow on this mesh size function
        timestep : float
            Timestep size (:math:`seconds`) to
        wave_amplitude : float, default=2
            Free surface elevation (:math:`meters`) from the reference
            (i.e. wave height)
        source_index : int or list of ints or None, default=None
            The index of entries from the input list argument
            of the constructor of collector size function. If `None`
            all input rasters are used.

        Returns
        -------
        None
        """

        self._applied = False

        # TODO: Validate conflicting constraints, right now last one wins
        if upper_bound is None and lower_bound is None:
            raise ValueError("Both upper and lower Courant bounds can NOT be None!")

        if source_index is not None and not isinstance(source_index, (tuple, list)):
            source_index = [source_index]

        if upper_bound is not None:
            self._constraint_info_coll.add(
                source_index,
                CourantNumConstraint(
                    value=upper_bound,
                    timestep=timestep,
                    wave_amplitude=wave_amplitude,
                    value_type='min'
                )
            )
        if lower_bound is not None:
            self._constraint_info_coll.add(
                source_index,
                CourantNumConstraint(
                    value=lower_bound,
                    timestep=timestep,
                    wave_amplitude=wave_amplitude,
                    value_type='min'
                )
            )


    def add_region_constraint(
            self,
            value: Union[float, npt.NDArray[np.float32]],
            shape: Union[Polygon, MultiPolygon],
            crs: Union[CRS, str] = None,
            value_type: Literal['min', 'max'] = 'min',
            rate: float = 0.01,
            source_index: Union[List[int], int, None] = None
            ) -> None:

        """Add a value contraint for the points in specified region

        Add a fixed-value constraint to the region of the size function
        specified by the input shape.  Optionally a `rate` can be
        specified to relax the constraint gradually outside the bounds.

        Parameters
        ----------
        value : float or array-like
            A single fixed value to be used for
            mesh size if condition is not met based on `value_type`.
        shape: MultiPolygon or Polygon
            Region of specified constraint
        crs: CRS or None, default=None
            The CRS of the input shape
        value_type : {'min', 'max'}, default='min'
            Type of contraint. If 'min', it means the mesh size
            should not be smaller than the specified `value` at each
            point.
        rate : float, default=0.01
            Rate of relaxation of constraint outside the specified region
        source_index: int or list of ints or None, default=None
            The index of input hfun source. If `None` all inputs
            are used.

        Returns
        -------
        None
        """
        self._constraint_info_coll.add(
            source_index,
            RegionConstraint(
                value=value,
                shape=shape,
                crs=crs,
                value_type=value_type,
                rate=rate
            )
        )


    def add_contour(
            self,
            level: Union[List[float], float, None] = None,
            expansion_rate: float = 0.01,
            target_size: Optional[float] = None,
            contour_defn: Optional[Contour] = None
            ) -> None:
        """Add refinement for auto extracted contours

        Add refinement for the contour lines extracted based on
        level or levels specified by `level`. The refinement
        is relaxed with `expansion_rate` and distance from the
        extracted contour lines. Contours are extracted only from the
        raster inputs, but are applied on all inputs passed to the
        collector constructor.

        Parameters
        ----------
        level : float or list of floats or None, default=None
            Level(s) at which contour lines should be extracted.
        expansion_rate : float, default=0.01
            Rate to use for expanding refinement with distance away
            from the extracted contours.
        target_size : float or None, default=None
            Target size to use on the extracted contours and expand
            from with distance.
        contour_defn : Contour or None, default=None
            Contour definition objects which defines contour extraction
            specification from rasters.

        Returns
        -------
        None
        """

        # Always lazy
        self._applied = False

        levels = []
        if isinstance(level, (list, tuple)):
            levels.extend(level)
        else:
            levels.append(level)


        contour_defns = []
        if contour_defn is None:
            for lvl in levels:
                contour_defns.append(Contour(level=lvl))

        elif not isinstance(contour_defn, Contour):
            raise TypeError(
                f"Contour definition must be of type {Contour} not"
                f" {type(contour_defn)}!")

        elif level is not None:
            msg = "Level is ignored since a contour definition is provided!"
            warnings.warn(msg)
            _logger.info(msg)

        else:
            contour_defns.append(contour_defn)

        for ctr_dfn in contour_defns:
            self._contour_info_coll.add(
                ctr_dfn,
                expansion_rate=expansion_rate,
                target_size=target_size)

    def add_channel(
            self,
            level: float = 0,
            width: float = 1000,
            target_size: float = 200,
            expansion_rate: Optional[float] = None,
            tolerance: Optional[float] = None,
            channel_defn: Optional[Channel] = None
            ) -> None:
        """Add refinement for auto detected channels

        Automatically detects narrow regions in the domain and apply
        refinement size with an expanion rate (if provided) outside
        the detected area.

        Parameters
        ----------
        level : float, default=0
            High water mark at which domain is extracted for narrow
            region or channel calculations.
        width : float, default=1000
            The cut-off width for channel detection.
        target_size : float, default=200
            Target size to use on the detected channels and expand
            from with distance.
        expansion_rate : float or None, default=None
            Rate to use for expanding refinement with distance away
            from the detected channels.
        tolerance : float or None, default=None
            Tolerance to use for simplifying the polygon extracted
            from DEM data. If `None` don't simplify.
        channel_defn : Contour or None
            Channel definition objects which defines channel
            extraction specification from rasters.

        Returns
        -------
        None
        """

        self._applied = False

        # Always lazy
        self._applied = False

        # Even a tolerance of 1 for simplifying polygon for channel
        # calculations is much faster than no simplification. 50
        # is much faster than 1. The reason is in simplify we don't
        # preserve topology
        if channel_defn is None:
            channel_defn = Channel(
                level=level, width=width, tolerance=tolerance)

        elif not isinstance(channel_defn, Channel):
            raise TypeError(
                f"Channel definition must be of type {Channel} not"
                f" {type(channel_defn)}!")

        self._ch_info_coll.add(
            channel_defn,
            expansion_rate=expansion_rate,
            target_size=target_size)


    def add_subtidal_flow_limiter(
            self,
            hmin: Optional[float] = None,
            hmax: Optional[float] = None,
            lower_bound: Optional[float] = None,
            upper_bound: Optional[float] = None,
            source_index: Union[List[int], int, None] = None
            ) -> None:
        """Add mesh refinement based on topography.

        Calculates a pre-defined function of topography to use
        as values for refinement. The function values are cut off
        using `hmin` and `hmax` and is applied to the region bounded
        by `lower_bound` and `upper_bound`.

        Parameters
        ----------
        hmin : float or None, default=None
            Minimum mesh size in the refinement
        hmax : float or None, default=None
            Maximum mesh size in the refinement
        lower_bound : float or None, default=None
            Lower limit of the cut-off elevation for region to apply
            the fixed `value`.
        upper_bound : float or None, default=None
            Higher limit of the cut-off elevation for region to apply
            the fixed `value`.
        source_index : int or list of ints or None, default=None
            The index of raster entries from the input list argument
            of the constructor of collector size function. If `None`
            all input rasters are used.

        Returns
        -------
        None

        See Also
        --------
        hfun.raster.HfunRaster.add_subtidal_flow_limiter :
        """

        self._applied = False

        if source_index is not None and not isinstance(source_index, (tuple, list)):
            source_index = [source_index]

        # TODO: Checks on hmin/hmax, etc?

        self._flow_lim_coll.add(
            source_index,
            hmin=hmin,
            hmax=hmax,
            upper_bound=upper_bound,
            lower_bound=lower_bound)


    def add_constant_value(
            self,
            value: float,
            lower_bound: Optional[float] = None,
            upper_bound: Optional[float] = None,
            source_index: Union[List[int], int, None] =None):
        """Add refinement of fixed value in the region specified by bounds.

        Apply fixed value mesh size refinement to the region specified
        by bounds (if provided) or the whole domain.

        Parameters
        ----------
        value : float
            Fixed value to use for refinement size
        lower_bound : float or None, default=None
            Lower limit of the cut-off elevation for region to apply
            the fixed `value`.
        upper_bound : float or None, default=None
            Higher limit of the cut-off elevation for region to apply
            the fixed `value`.
        source_index : int or list of ints or None, default=None
            The index of raster entries from the input list argument
            of the constructor of collector size function. If `None`
            all input rasters are used.

        Returns
        -------
        None
        """

        self._applied = False

        contour_defn0 = None
        contour_defn1 = None
        if lower_bound is not None and not np.isinf(lower_bound):
            contour_defn0 = Contour(level=lower_bound)
        if upper_bound is not None and not np.isinf(upper_bound):
            contour_defn1 = Contour(level=upper_bound)

        if source_index is not None and not isinstance(source_index, (tuple, list)):
            source_index = [source_index]
        self._const_val_contour_coll.add(
            source_index, contour_defn0, contour_defn1, value)


    def add_patch(
            self,
            shape: Union[MultiPolygon, Polygon, None] = None,
            patch_defn: Optional[Patch] = None,
            shapefile: Union[None, str, Path] = None,
            expansion_rate: Optional[float] = None,
            target_size: Optional[float] = None,
            ) -> None:
        """Add refinement as a region of fixed size with an optional rate

        Add a refinement based on a region specified by `shape`,
        `patch_defn` or `shapefile`.  The fixed `target_size`
        refinement can be expanded outside the region specified by the
        shape if `expansion_rate` is provided.

        Parameters
        ----------
        shape : MultiPolygon or Polygon or None, default=None
            Shape of the region to use specified `target_size` for
            refinement. Only one of `shape`, `patch_defn` or `shapefile`
            must be specified.
        patch_defn : Patch or None, default=None
            Shape of the region to use specified `target_size` for
            refinement. Only one of `shape`, `patch_defn` or `shapefile`
            must be specified.
        shapefile : None or str or Path, default=None
            Shape of the region to use specified `target_size` for
            refinement. Only one of `shape`, `patch_defn` or `shapefile`
            must be specified.
        expansion_rate : float or None, default=None
            Optional rate to use for expanding refinement outside
            the specified shape in `multipolygon`.
        target_size : float or None, default=None
            Fixed target size of mesh to use for refinement in
            `multipolygon`

        Returns
        -------
        None
        """

        # "shape" should be in 4326 CRS. For shapefile or patch_defn
        # CRS info is included

        self._applied = False

        if not patch_defn:
            if shape:
                patch_defn = Patch(shape=shape)

            elif shapefile:
                patch_defn = Patch(shapefile=shapefile)

        self._refine_patch_info_coll.add(
            patch_defn,
            expansion_rate=expansion_rate,
            target_size=target_size)

    def add_feature(
            self,
            shape: Union[MultiLineString, LineString, None] = None,
            line_defn: Optional[LineFeature] = None,
            shapefile: Union[None, str, Path] = None,
            expansion_rate: float = 0.01,
            target_size: Optional[float] = None,
            crs: CRS = 4326
            ) -> None:
        """Add refinement as a region of fixed size with an optional rate

        Add a refinement based on lines specified by `shape`,
        `line_defn` or `shapefile`.  The fixed `target_size`
        refinement is expanded by the `expansion_rate`.

        Parameters
        ----------
        shape : MultiLineString or LineString or None, default=None
            Shape of the region to use specified `target_size` for
            refinement. Only one of `shape`, `line_defn` or `shapefile`
            must be specified.
        line_defn : LineFeature or None, default=None
            Shape of the region to use specified `target_size` for
            refinement. Only one of `shape`, `line_defn` or `shapefile`
            must be specified.
        shapefile : None or str or Path, default=None
            Shape of the region to use specified `target_size` for
            refinement. Only one of `shape`, `line_defn` or `shapefile`
            must be specified.
        expansion_rate : float, default=0.01
            Rate to use for expanding refinement away from
            the specified shape
        target_size : float or None, default=None
            Fixed target size of mesh to use for refinement in
            `multipolygon`
        crs : CRS, default 4326
            The CRS of the input `shape`

        Returns
        -------
        None
        """

        self._applied = False

        if not line_defn:
            if shape:
                line_defn = LineFeature(shape=shape, shape_crs=crs)

            elif shapefile:
                line_defn = LineFeature(shapefile=shapefile)

        self._refine_line_info_coll.add(
            line_defn,
            expansion_rate=expansion_rate,
            target_size=target_size)


    @staticmethod
    def _type_chk(input_list: List[Any]) -> None:
        """Checks the if the input types are supported for size function

        Checks if size function collector supports handling
        size functions created from the input types.

        Parameters
        ----------
        input_list : List[Any]

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If the any of the inputs are not supported
        """

        valid_types = (str, Path, Raster, BaseMesh, HfunRaster, HfunMesh)
        if not all(isinstance(item, valid_types) for item in input_list):
            raise TypeError(
                f'Input list items must be of type'
                f' {", ".join(str(i) for i in valid_types)},'
                f' or a derived type.')

    def _apply_features(self) -> None:
        """Internal: apply all specified refinements and constraints

        Apply all specified refinements and constrains for the exact
        algorithm.

        Parameters
        ----------

        Returns
        -------
        None

        See Also
        --------
        _apply_features_fast :
        """

        if not self._applied:
            self._apply_contours()
            self._apply_flow_limiters()
            self._apply_const_val()
            self._apply_linefeatures()
            self._apply_patch()
            self._apply_channels()
            self._apply_constraints()

        self._applied = True


    def _apply_constraints(self) -> None:
        """Internal: apply specified constraints.

        Apply specified constraints for the exact algorithm.

        Parameters
        ----------

        Returns
        -------
        None

        See Also
        --------
        _apply_constraints_fast :
        """

        if self._method == 'fast':
            raise NotImplementedError(
                "This function does not suuport fast hfun method")

        self._constraint_info_coll.apply(self._hfun_list)


    def _apply_contours(self, apply_to: Optional[SizeFuncList] = None) -> None:
        """Internal: apply specified constraints.

        Parameters
        ----------
        apply_to : SizeFuncList or None, default=None
            Size functions on which contours must be applied. If `None`
            all inputs are used to apply the calculated contours.

        Returns
        -------
        None
        """

        # TODO: Consider CRS before applying to different hfuns
        #
        # NOTE: for parallelization make sure a single hfun is NOT
        # passed to multiple processes

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]
        if apply_to is None:
            mesh_hfun_list = [
                i for i in self._hfun_list if isinstance(i, HfunMesh)]
            if self._base_mesh and self._base_as_hfun:
                mesh_hfun_list.insert(0, self._base_mesh)
            apply_to = [*mesh_hfun_list, *raster_hfun_list]

        with tempfile.TemporaryDirectory() as temp_path:
            with Pool(processes=self._nprocs) as p:
                # Contours are ONLY extracted from raster sources
                self._contour_coll.calculate(raster_hfun_list, temp_path)
                counter = 0
                for hfun in apply_to:
                    for gdf in self._contour_coll:
                        for row in gdf.itertuples():
                            _logger.debug(row)
                            shape = row.geometry
                            if isinstance(shape, GeometryCollection):
                                continue
                            # NOTE: CRS check is done AFTER
                            # GeometryCollection check because
                            # gdf.to_crs results in an error in case
                            # of empty GeometryCollection
                            if not gdf.crs.equals(hfun.crs):
                                _logger.info("Reprojecting feature...")
                                transformer = Transformer.from_crs(
                                    gdf.crs, hfun.crs, always_xy=True)
                                shape = ops.transform(
                                        transformer.transform, shape)
                            counter = counter + 1
                            hfun.add_feature(**{
                                'feature': shape,
                                'expansion_rate': row.expansion_rate,
                                'target_size': row.target_size,
                                'pool': p
                            })
            p.join()

            # hfun objects cause issue with pickling
            # -> cannot be passed to pool
#            with Pool(processes=self._nprocs) as p:
#                p.starmap(
#                    _apply_contours_worker,
#                    [(hfun, self._contour_coll, self._nprocs)
#                     for hfun in apply_to])

    def _apply_channels(self, apply_to: Optional[SizeFuncList] = None) -> None:
        """Internal: apply specified channel refinements.

        Parameters
        ----------
        apply_to : SizeFuncList or None, default=None
            Size functions on which channels must be applied. If `None`
            all inputs are used to apply the calculated channels.

        Returns
        -------
        None
        """

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]
        if apply_to is None:
            mesh_hfun_list = [
                i for i in self._hfun_list if isinstance(i, HfunMesh)]
            if self._base_mesh and self._base_as_hfun:
                mesh_hfun_list.insert(0, self._base_mesh)
            apply_to = [*mesh_hfun_list, *raster_hfun_list]

        with tempfile.TemporaryDirectory() as temp_path:
            # Channels are ONLY extracted from raster sources
            self._channels_coll.calculate(raster_hfun_list, temp_path)
            counter = 0
            for hfun in apply_to:
                for gdf in self._channels_coll:
                    for row in gdf.itertuples():
                        _logger.debug(row)
                        shape = row.geometry
                        if isinstance(shape, GeometryCollection):
                            continue
                        # NOTE: CRS check is done AFTER
                        # GeometryCollection check because
                        # gdf.to_crs results in an error in case
                        # of empty GeometryCollection
                        if not gdf.crs.equals(hfun.crs):
                            _logger.info("Reprojecting feature...")
                            transformer = Transformer.from_crs(
                                gdf.crs, hfun.crs, always_xy=True)
                            shape = ops.transform(
                                    transformer.transform, shape)
                        counter = counter + 1
                        hfun.add_patch(**{
                            'multipolygon': shape,
                            'expansion_rate': row.expansion_rate,
                            'target_size': row.target_size,
                            'nprocs': self._nprocs
                        })


    @property
    def execution_mode(self) -> str:
        """
        Gets the current execution mode for refinements ('serial' or 'parallel').
        Defaults to 'serial' if not previously set.
        """
        # Lazy Initialization: If the attribute doesn't exist yet,
        # create it here with the default value.
        if not hasattr(self, '_execution_mode'):
            self._execution_mode = 'serial'
        return self._execution_mode


    @execution_mode.setter
    def execution_mode(self, mode: str) -> None:
        """
        Sets the execution mode for refinements.

        Parameters
        ----------
        mode : str
            The desired mode. Must be either 'serial' or 'parallel'.
        """
        if mode not in ['serial', 'parallel']:
            raise ValueError("Execution must be either 'serial' or 'parallel'")

        if mode == 'parallel' and (self._nprocs is None or self._nprocs <= 1):
            warnings.warn(
                f"Execution mode set to 'parallel' but nprocs is {self._nprocs}. "
            "Execution will fall back to serial. Set nprocs > 1 for parallel."
            )

        self._execution_mode = mode


    def _apply_flow_limiters(self) -> None:
        """
        Dispatches to either the serial or parallel implementation based on
        the current execution mode.
        """
        if self.execution_mode == 'parallel' and self._nprocs > 1:
            _logger.info("Applying flow limiters using PARALLEL method.")
            self._apply_flow_limiters_parallel()
        else:
            _logger.info("Applying flow limiters using SERIAL method.")
            self._apply_flow_limiters_serial()


    def _apply_flow_limiters_serial(self) -> None:
        """Internal: apply specified sub tidal flow limiter refinements

        Applies specified subtidal flow limiter refinements for
        the exact algorithm.

        Parameters
        ----------

        Returns
        -------
        None

        See Also
        --------
        _apply_flow_limiters_fast :
        """

        if self._method == 'fast':
            raise NotImplementedError(
                "This function does not suuport fast hfun method")

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]

        for in_idx, hfun in enumerate(raster_hfun_list):
            for src_idx, hmin, hmax, zmax, zmin in self._flow_lim_coll:
                if src_idx is not None and in_idx not in src_idx:
                    continue
                if hmin is None:
                    hmin = self._size_info['hmin']
                if hmax is None:
                    hmax = self._size_info['hmax']
                hfun.add_subtidal_flow_limiter(hmin, hmax, zmin, zmax)


    def _apply_flow_limiters_parallel(self) -> None:
        """
        Applies specified subtidal flow limiter refinements in parallel using a
        robust file-based worker pattern.

        This method coordinates the parallel processing by:
        1.  Preparing simple, pickleable 'task' dictionaries for each HfunRaster
            that needs modification.
        2.  Distributing these tasks to a pool of worker processes.
        3.  Integrating the results by creating new HfunRaster objects from the
            output files generated by the workers.
        """
        if self._method == 'fast':
            raise NotImplementedError(
            "This function is part of the 'exact' method and doesnt support 'fast' mode."
            )

        # Phase 1: PREPARATION (Coordinator)
        # This phase gathers all the work that needs to be done and packages it
        # into a list of simple, pickleable tasks for the worker processes.

        tasks = []
        hfuns_to_process = {}

        # First, group all applicable refinement rules by the HfunRaster they apply to.
        # This is more efficient than creating a separate task for every single rule.

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]

        for in_idx, hfun in enumerate(raster_hfun_list):
            limiter_rules_for_this_hfun = []
            for src_idx, hmin, hmax, zmax, zmin in self._flow_lim_coll:
                # Check if the rule applies to this specific hfun instance
                if src_idx is None or in_idx in src_idx:
                    limiter_rules_for_this_hfun.append({
                    'hmin': hmin if hmin is not None else self._size_info.get('hmin'),
                    'hmax': hmax if hmax is not None else self._size_info.get('hmax'),
                    'zmin': zmin,
                    'zmax': zmax
                    })

            # If any rules were found, mark this HfunRaster for processing.
            if limiter_rules_for_this_hfun:
                hfuns_to_process[in_idx] = {
                    'hfun': hfun,
                    'rules': limiter_rules_for_this_hfun
                }

        # Now, create the simple task dictionaries that can be sent to the pool.
        for in_idx, data in hfuns_to_process.items():
            hfun = data['hfun']

            # Determine the correct input file. If this raster was already processed
            # by another step (e.g., _apply_constraints), use that output file.
            # Otherwise, use the original hfun path.
            hfun_input_path = hfun.tmpfile


            # The path to the original, unmodified topography/DEM data.
            topo_input_path = hfun._raster.path # pylint: disable=W0212

            # Define a unique output path in our persistent working directory.
            output_path = os.path.join(self._work_dir,
                                       f"flow_limiter_result_{in_idx}.tif")

            task = {
                'original_index': in_idx,
                'hfun_input_path': hfun_input_path,
                'topo_input_path': topo_input_path,
                'output_path': output_path,
                'global_hmin': hfun._hmin, # pylint: disable=W0212
                'global_hmax': hfun._hmax, # pylint: disable=W0212
                'limiter_params': data['rules']
            }
            tasks.append(task)


        # If no tasks were generated, there's nothing to do.
        if not tasks:
            _logger.info("No flow limiter tasks to execute.")
            return

        # Phase 2: EXECUTION (Distribute to Laborers)
        # This phase sends the prepared tasks to a pool of worker processes
        # and waits for them to complete the heavy computational work.

        _logger.info(f"Start parallel execution for {len(tasks)} flow limiter tasks")
        with Pool(processes=self._nprocs) as p:
            results = p.map(_flow_limiter_task_worker, tasks)
        _logger.info("Parallel execution finished.")

        # Phase 3: INTEGRATION (Process Results)
        # This phase takes the results from the workers (which are just file paths)
        # and updates the main HfunCollector's state with the new, processed data.

        new_hfun_objects = {}
        for result in results:
            if result['status'] == 'error':
                _logger.error(
                    f"Worker failed for HfunRaster at index "
                    f"{result['original_index']}: {result['error']}"
                )
                continue

            idx = result['original_index']
            output_path = result['output_path']
            original_hfun = self._hfun_list[idx]
            # Create a new, updated HfunRaster instance to replace the old one.
            new_hfun_objects[idx] = HfunRaster(
                raster=original_hfun.raster,
                hmin=original_hfun.hmin,
                hmax=original_hfun.hmax,
                verbosity=original_hfun.verbosity,
                initial_value=output_path
            )

        # Finally, update the main list with the new objects.
        for idx, new_hfun in new_hfun_objects.items():
            _logger.info(f"Updating HfunCol. state with processed raster for idx {idx}.")
            self._hfun_list[idx] = new_hfun


    def _apply_const_val(self) -> None:
        """
        Dispatches to either the serial or parallel implementation based on
        the current execution mode.
        """
        if self.execution_mode == 'parallel' and self._nprocs > 1:
            _logger.info("Applying constant values using PARALLEL method.")
            self._apply_const_val_parallel()
        else:
            _logger.info("Applying constant values using SERIAL method.")
            self._apply_const_val_serial()


    def _apply_const_val_serial(self):
        """Internal: apply specified constant value refinements.

        Applies constant value refinements for the exact algorithm.

        Returns
        -------
        None

        See Also
        --------
        _apply_const_val_fast :
        """

        if self._method == 'fast':
            raise NotImplementedError(
                "This function does not suuport fast hfun method")

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]

        for in_idx, hfun in enumerate(raster_hfun_list):
            for (src_idx, ctr0, ctr1), const_val in self._const_val_contour_coll:
                if src_idx is not None and in_idx not in src_idx:
                    continue
                level0 = None
                level1 =  None
                if ctr0 is not None:
                    level0 = ctr0.level
                if ctr1 is not None:
                    level1 = ctr1.level
                hfun.add_constant_value(const_val, level0, level1)


    def _apply_const_val_parallel(self):
        """Internal: apply specified constant value refinements.

        Applies constant value refinements for the exact algorithm.

        Returns
        -------
        None

        See Also
        --------
        _apply_const_val_fast :
        """

        if self._method == 'fast':
            raise NotImplementedError(
                "This function does not suuport fast hfun method")

       # Phase 1: PREPARATION (Coordinator)
        tasks = []
        hfuns_to_process = {}

        # Group all applicable constant value rules by the HfunRaster they apply to.

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]

        for in_idx, hfun in enumerate(raster_hfun_list):
            rules_for_this_hfun = []
            for (src_idx, ctr0, ctr1), const_val in self._const_val_contour_coll:
                if src_idx is None or in_idx in src_idx:
                    rules_for_this_hfun.append({
                        'value': const_val,
                        'lower_bound': ctr0.level if ctr0 else None,
                        'upper_bound': ctr1.level if ctr1 else None
                    })

            if rules_for_this_hfun:
                hfuns_to_process[in_idx] = { 'hfun': hfun,
                                            'rules': rules_for_this_hfun }

        # Now, create the simple task dictionaries for the pool.
        for in_idx, data in hfuns_to_process.items():
            hfun = data['hfun']
            # Determine the correct input file.If this raster was already processed
            hfun_input_path = hfun.tmpfile
            topo_input_path = hfun._raster.path # pylint: disable=W0212

            output_path = os.path.join(self._work_dir,
                                       f"const_val_result_{in_idx}.tif")

            task = {
                'original_index': in_idx,
                'hfun_input_path': hfun_input_path,
                'topo_input_path': topo_input_path,
                'output_path': output_path,
                'global_hmin': hfun._hmin, # pylint: disable=W0212
                'global_hmax': hfun._hmax, # pylint: disable=W0212
                'const_val_rules': data['rules']
            }
            tasks.append(task)

        if not tasks:
            _logger.info("No constant value tasks to execute.")
            return


            # Phase 2: EXECUTION
        _logger.info(f"Start parallel execution for {len(tasks)} const. value tasks")
        with Pool(processes=self._nprocs) as p:
            results = p.map(_const_val_task_worker, tasks)
        _logger.info("Parallel execution finished.")

        # Phase 3: INTEGRATION
        new_hfun_objects = {}
        for result in results:
            if result['status'] == 'error':
                _logger.error("Val worker %s failed: %s",
                              result['original_index'],
                              result['error'])
                continue

            idx = result['original_index']
            output_path = result['output_path']
            original_hfun = self._hfun_list[idx]

            # Create the new HfunRaster, ensuring we don't wipe the worker's data.
            new_hfun_objects[idx] = HfunRaster(
            raster=original_hfun.raster, # Pass the original topography Raster object
            hmin=original_hfun.hmin,
            hmax=original_hfun.hmax,
            verbosity=original_hfun.verbosity,
            initial_value=output_path # Pass the path to the file created by the worker
            )

        # Finally, update the main list with the newly created objects.
        for idx, new_hfun in new_hfun_objects.items():
            _logger.info(f"Update HfunCollector with const_val raster at idx {idx}.")
            self._hfun_list[idx] = new_hfun


    def _apply_patch(self, apply_to: Optional[SizeFuncList] = None) -> None:
        """Internal: apply the specified patch refinements.

        Parameters
        ----------
        apply_to : SizeFuncList or None, default=None
            Size functions on which patches must be applied. If `None`
            all inputs are used to apply the patches.

        Returns
        -------
        None
        """

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]
        if apply_to is None:
            mesh_hfun_list = [
                i for i in self._hfun_list if isinstance(i, HfunMesh)]
            if self._base_mesh and self._base_as_hfun:
                mesh_hfun_list.insert(0, self._base_mesh)
            apply_to = [*mesh_hfun_list, *raster_hfun_list]

        # TODO: Parallelize
        for hfun in apply_to:
            for patch_defn, size_info in self._refine_patch_info_coll:
                shape, crs = patch_defn.get_multipolygon()
                if hfun.crs != crs:
                    transformer = Transformer.from_crs(
                        crs, hfun.crs, always_xy=True)
                    shape = ops.transform(
                            transformer.transform, shape)

                hfun.add_patch(
                        shape, nprocs=self._nprocs, **size_info)


    def _apply_linefeatures(self, apply_to: Optional[SizeFuncList] = None) -> None:
        """Internal: apply the specified line feature refinements.

        Parameters
        ----------
        apply_to : SizeFuncList or None, default=None
            Size functions on which line features must be applied. If `None`
            all inputs are used to apply the line features.

        Returns
        -------
        None
        """

        raster_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]
        if apply_to is None:
            mesh_hfun_list = [
                i for i in self._hfun_list if isinstance(i, HfunMesh)]
            if self._base_mesh and self._base_as_hfun:
                mesh_hfun_list.insert(0, self._base_mesh)
            apply_to = [*mesh_hfun_list, *raster_hfun_list]

        # TODO: Parallelize
        with Pool(processes=self._nprocs) as p:
            for hfun in apply_to:
                for lineftr_defn, size_info in self._refine_line_info_coll:
                    shape, crs = lineftr_defn.get_multiline()
                    if hfun.crs != crs:
                        transformer = Transformer.from_crs(
                            crs, hfun.crs, always_xy=True)
                        shape = ops.transform(
                                transformer.transform, shape)

                    hfun.add_feature(
                        feature=shape,
                        pool=p,
                        **size_info
                    )


    def _write_hfun_to_disk(
            self,
            out_path: Union[str, Path],
            **kwargs
            ) -> List[Union[str, Path]]:
        """Internal: write individual size function output mesh to file

        Calculate the interpolated on-mesh size function from each
        individual input, clip overlaps based on priority, and
        write the results to disk for later combining.

        Parameters
        ----------
        out_path : path-like
            The path of the (temporary) directory to which mesh size
            functions must be written.
        **kwargs : dict
            Arguments to pass to the hfun.meshdata() method (e.g. stride).

        Returns
        -------
        list of path-like
            List of individual file path for mesh size function of
            each input.
        """

        out_dir = Path(out_path)
        path_list = []
        file_counter = 0
        pid = os.getpid()
        bbox_list = []

        hfun_list = self._hfun_list[::-1]
        if self._base_mesh and self._base_as_hfun:
            hfun_list = [*self._hfun_list[::-1], self._base_mesh]

        # Last user input item has the highest priority (its trias
        # are not dropped) so process in reverse order
        for hfun in hfun_list:
            # TODO: Calling meshdata() on HfunMesh more than once causes
            # issue right now due to change in crs of internal Mesh
            try:
                meshdata_hfun = deepcopy(hfun.meshdata(**kwargs))
            except TypeError:
            # If a specific hfun type (like HfunMesh) doesn't support the args,
            # ignore them
            # This protects against passing 'stride' to a Mesh-based hfun which
            # might not need it
                meshdata_hfun = deepcopy(hfun.meshdata())

            # If no CRS info, we assume EPSG:4326
            if hasattr(meshdata_hfun, "crs"):
                dst_crs = CRS.from_user_input("EPSG:4326")
                if meshdata_hfun.crs != dst_crs:
                    utils.reproject(meshdata_hfun, dst_crs)

            # Get all previous bbox and clip to resolve overlaps
            # removing all tria that have NODE in bbox because it's
            # faster and so we can resolve all overlaps
            _logger.info("Removing bounds from hfun mesh...")
            for ibox in bbox_list:
                meshdata_hfun = utils.clip_mesh_by_shape(
                    meshdata_hfun,
                    ibox,
                    use_box_only=True,
                    fit_inside=True,
                    inverse=True)

            if len(meshdata_hfun.coords) == 0:
                _logger.debug("Hfun ignored due to overlap")
                continue

            # Check meshdata_hfun.value against hmin & hmax
            hmin = self._size_info['hmin']
            hmax = self._size_info['hmax']
            if hmin:
                meshdata_hfun.values[meshdata_hfun.values < hmin] = hmin
            if hmax:
                meshdata_hfun.values[meshdata_hfun.values > hmax] = hmax

            mesh = Mesh(meshdata_hfun)
            bbox_list.append(mesh.get_bbox(crs="EPSG:4326"))
            file_counter = file_counter + 1
            _logger.info(f'write mesh {file_counter} to file...')
            file_path = out_dir / f'hfun_{pid}_{file_counter}.2dm'
            mesh.write(file_path, format='2dm')
            path_list.append(file_path)
            _logger.info('Done writing 2dm file.')
            del mesh
            gc.collect()
        return path_list



    def _get_hfun_composite(
            self,
            hfun_path_list: List[Union[str, Path]]
            ) -> MeshData:
        """Internal: combine the size functions written to disk

        Combine the size functions written to disk from the list of
        input files `hfun_path_list`. This is used for `exact` method.

        Parameters
        ----------
        hfun_path_list : list of path-like

        Retruns
        -------
        MeshData
            The combined size function interpolated on an optimized
            mesh.

        See Also
        --------
        _get_hfun_composite_fast :
        """

        collection = []
        _logger.info('Reading 2dm hfun files...')
        start = time()
        for path in hfun_path_list:
            collection.append(Mesh.open(path, crs='EPSG:4326'))
        _logger.info(f'Reading 2dm hfun files took {time()-start}.')

        # NOTE: Overlaps are taken care of in the write stage

        coord = []
        index = []
        value = []
        offset = 0
        for hfun in collection:
            index.append(hfun.triangles + offset)
            coord.append(hfun.coord)
            value.append(hfun.value.reshape(-1, 1))
            offset += hfun.coord.shape[0]

        composite_hfun = MeshData(
                coords=np.vstack(coord),
                tria=np.vstack(index),
                values=np.vstack(value),
                crs=CRS.from_user_input("EPSG:4326")
        )

        # NOTE: In the end we need to return in a CRS that
        # uses meters as units. UTM based on the center of
        # the bounding box of the hfun is used
        # Up until now all calculation was in EPSG:4326
        utils.project_to_utm(composite_hfun)

        return composite_hfun


    def _create_big_raster(self, out_path: Union[str, Path]) -> Raster:
        """Internal: create a large raster covering all input rasters.

        Parameters
        ----------
        out_path : path-like
            Path of the (tempoerary) directory to which the large
            raster needs to be written

        Returns
        -------
        Raster
            Lower resolution raster covering all input rasters.
        """

        out_dir = Path(out_path)
        out_rast = out_dir / 'big_raster.tif'

        rast_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]
        if len(rast_hfun_list) == 0:
            return None

        all_bounds = []
        n_cell_lim = 0
        for hfun_in in rast_hfun_list:
            n_cell_lim = max(
                hfun_in.raster.src.shape[0]
                    * hfun_in.raster.src.shape[1],
                n_cell_lim)
            all_bounds.append(
                    hfun_in.get_bbox(crs='EPSG:4326').bounds)
        # 3 is just a arbitray tolerance for memory limit calculations
        n_cell_lim = n_cell_lim * self._nprocs / 3
        all_bounds = np.array(all_bounds)

        x0, y0 = np.min(all_bounds[:, [0, 1]], axis=0)
        x1, y1 = np.max(all_bounds[:, [2, 3]], axis=0)

        utm_crs = utils.estimate_bounds_utm(
                (x0, y0, x1, y1), "EPSG:4326")
        assert utm_crs is not None
        transformer = Transformer.from_crs(
                'EPSG:4326', utm_crs, always_xy=True)

        # If it's the full earth, then the spaces are at least 1/2 deg
        xs = np.linspace(x0, x1, 720)
        ys = np.linspace(y0, y1, 720)
        coords = [[x0, y] for y in ys]
        coords.extend([[x, y1] for x in xs])
        coords.extend([[x1, y] for y in reversed(ys)])
        coords.extend([[x, y0] for x in reversed(xs)])
        poly_epsg4326 = Polygon(np.array(coords))
        poly_utm = ops.transform(transformer.transform, poly_epsg4326)
        x0, y0, x1, y1 = poly_utm.bounds

        worst_res = 0
        for hfun_in in rast_hfun_list:
            bnd1 = hfun_in.get_bbox(crs=utm_crs).bounds
            dim1 = np.max([bnd1[2] - bnd1[0], bnd1[3] - bnd1[1]])
            bnd2 = hfun_in.get_bbox(crs='EPSG:4326').bounds
            dim2 = np.max([bnd2[2] - bnd2[0], bnd2[3] - bnd2[1]])
            ratio = dim1 / dim2
            pixel_size_x = hfun_in.raster.src.transform[0] * ratio
            pixel_size_y = -hfun_in.raster.src.transform[4] * ratio

            worst_res = np.max([worst_res, pixel_size_x, pixel_size_y])

        # TODO: What if no hmin? -> use smallest raster res!
        g_hmin = self._size_info['hmin']
        res = np.max([g_hmin / 2, worst_res])
        _logger.info(
                f"Spatial resolution"
                f" chosen: {res}, worst: {worst_res}")
        shape0 = int(np.ceil(abs(x1 - x0) / res))
        shape1 = int(np.ceil(abs(y1 - y0) / res))

        approx =  int(np.sqrt(n_cell_lim))
        window_size = None #default of OCSMesh.raster.Raster
        mem_lim = 0 # default of rasterio
        if approx < max(shape0, shape1):
            window_size = np.min([shape0, shape1, approx])
            # Memory limit in MB
            mem_lim = n_cell_lim * np.float32(1).itemsize / 10e6


        # NOTE: Upper-left vs lower-left origin
        # (this only works for upper-left)
        transform = from_origin(x0 - res / 2, y1 + res / 2, res, res)

        rast_profile = {
                'driver': 'GTiff',
                'dtype': np.float32,
                'width': shape0,
                'height': shape1,
                'crs': utm_crs,
                'transform': transform,
                'count': 1,
        }
        with rasterio.open(str(out_rast), 'w', **rast_profile) as dst:
            # For places where raster is DEM is not provided it's
            # assumed deep ocean for contouring purposes
            if window_size is not None:
                write_wins = get_iter_windows(
                    shape0, shape1, chunk_size=window_size)
                for win in write_wins:
                    z = np.full((win.width, win.height), -99999, dtype=np.float32)
                    dst.write(z, 1, window=win)

            else:
                z = np.full((shape0, shape1), -99999, dtype=np.float32)
                dst.write(z, 1)
                del z


            # Reproject if needed (for now only needed if constant
            # value levels or subtidal limiters are added)
            for in_idx, hfun in enumerate(rast_hfun_list):
                ignore = True
                for (src_idx, _, _), _ in self._const_val_contour_coll:
                    if src_idx is None or in_idx in src_idx:
                        ignore = False
                        break
                for src_idx, _, _, _, _ in self._flow_lim_coll:
                    if src_idx is None or in_idx in src_idx:
                        ignore = False
                        break
                for src_idx, _ in self._constraint_info_coll:
                    if src_idx is None or in_idx in src_idx:
                        ignore = False
                        break
                if ignore:
                    continue

                # NOTE: Last one implicitely has highest priority in
                # case of overlap
                reproject(
                    source=rasterio.band(hfun.raster.src, 1),
                    destination=rasterio.band(dst, 1),
                    resampling=Resampling.nearest,
                    init_dest_nodata=False, # To avoid overwrite
                    num_threads=self._nprocs,
                    warp_mem_limit=mem_lim)



        return Raster(out_rast, chunk_size=window_size)

    def _apply_features_fast(self, big_raster: HfunRaster):
        """Internal: apply all specified refinements and constraints

        Apply all specified refinements and constrains for the fast
        algorithm.

        Parameters
        ----------
        big_raster : HfunRaster
            The lower-resolution large raster that covers all the
            input rasters.

        Returns
        -------
        None

        See Also
        --------
        _apply_features :
        """

        # NOTE: Caching applied doesn't work here since we apply
        # everything on a temporary big raster
        rast_hfun_list = []
        hfun_rast = None
        if big_raster:
            hfun_rast = HfunRaster(big_raster, **self._size_info)
            rast_hfun_list.append(hfun_rast)



        mesh_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunMesh)]
        if self._base_mesh and self._base_as_hfun:
            mesh_hfun_list.insert(0, self._base_mesh)

        # Mesh hfun parts are still stateful
        self._apply_contours([*mesh_hfun_list, *rast_hfun_list])
        if hfun_rast:
            # In fast method we only have big raster if any
            self._apply_flow_limiters_fast(hfun_rast)
            self._apply_const_val_fast(hfun_rast)
        # Mesh hfun parts are still stateful
        self._apply_linefeatures([*mesh_hfun_list, *rast_hfun_list])
        self._apply_patch([*mesh_hfun_list, *rast_hfun_list])
        self._apply_channels([*mesh_hfun_list, *rast_hfun_list])

        if hfun_rast:
            self._apply_constraints_fast(hfun_rast)


        return hfun_rast

    def _apply_flow_limiters_fast(self, big_hfun: HfunRaster) -> None:
        """Internal: apply specified sub tidal flow limiter refinements

        Applies specified subtidal flow limiter refinements for
        the fast algorithm.

        Parameters
        ----------
        big_hfun : HfunRaster
            The lower-resolution large raster that covers all the
            input rasters.

        Returns
        -------
        None

        See Also
        --------
        _apply_flow_limiters :
        """

        for src_idx, hmin, hmax, zmax, zmin in self._flow_lim_coll:
            # TODO: Account for source index
            if hmin is None:
                hmin = self._size_info['hmin']
            if hmax is None:
                hmax = self._size_info['hmax']

            # To avoid sharp gradient where no raster is projected
            if zmin is None:
                zmin = -99990
            else:
                zmin = max(zmin, -99990)

            big_hfun.add_subtidal_flow_limiter(hmin, hmax, zmin, zmax)

    def _apply_const_val_fast(self, big_hfun):
        """Internal: apply specified constant value refinements.

        Applies constant value refinements for the fast algorithm.

        Returns
        -------
        None

        See Also
        --------
        _apply_const_val :
        """

        for (src_idx, ctr0, ctr1), const_val in self._const_val_contour_coll:
            # TODO: Account for source index
            level0 = None
            level1 =  None
            if ctr0 is not None:
                level0 = ctr0.level
            if ctr1 is not None:
                level1 = ctr1.level
            big_hfun.add_constant_value(const_val, level0, level1)


    def _apply_constraints_fast(self, big_hfun: HfunRaster) -> None:
        """Internal: apply specified constraints.

        Apply specified constraints for the fast algorithm.

        Parameters
        ----------

        Returns
        -------
        None

        See Also
        --------
        _apply_constraints
        """

        # TODO: Account for source index
        self._constraint_info_coll.apply([big_hfun], per_hfun=False)


    def _get_hfun_composite_fast(self, big_hfun, **kwargs) -> MeshData:
        """Internal: combine the size function functions for fast method

        Combine the size functions of the large raster with non-raster
        inputs. This is used for `fast` method.

        Parameters
        ----------
        big_hfun : HfunRaster
            The single large raster based size function covering all
            input rasters.

        Retruns
        -------
        MeshData
            The combined size function interpolated on an optimized
            mesh.

        See Also
        --------
        _get_hfun_composite
        """

        # In fast method all DEM hfuns have more priority than all other inputs
        dem_hfun_list = [
            i for i in self._hfun_list if isinstance(i, HfunRaster)]
        nondem_hfun_list = [
            i for i in self._hfun_list if not isinstance(i, HfunRaster)]

        epsg4326 = CRS.from_user_input("EPSG:4326")

        dem_box_list = []
        for hfun in dem_hfun_list:
            dem_box_list.append(hfun.get_bbox(crs=epsg4326))

        index = []
        coord = []
        value = []
        offset = 0

        # Calculate multipoly and clip big hfun
        big_cut_shape = None
        if big_hfun:
            dem_gdf = gpd.GeoDataFrame(
                    geometry=dem_box_list, crs=epsg4326)
            big_cut_shape = dem_gdf.union_all()

            #  (e.g. mesh_engine='gmsh')
            big_meshdata = big_hfun.meshdata(**kwargs)

            if big_meshdata.crs is not None:
                if not epsg4326.equals(big_meshdata.crs):
                    utils.reproject(big_meshdata, epsg4326)

            big_meshdata = utils.clip_mesh_by_shape(
                big_meshdata,
                big_cut_shape,
                use_box_only=False,
                fit_inside=False)

            if big_meshdata.tria is not None:
                index.append(big_meshdata.tria + offset)

            coord.append(big_meshdata.coords)
            value.append(big_meshdata.values)
            offset = offset + coord[-1].shape[0]

        hfun_list = nondem_hfun_list[::-1]
        if self._base_mesh and self._base_as_hfun:
            hfun_list = [*nondem_hfun_list[::-1], self._base_mesh]

        nondem_shape_list = []
        for hfun in hfun_list:
            nondem_meshdata = deepcopy(hfun.meshdata())
            if hasattr(nondem_meshdata, "crs"):
                if not epsg4326.equals(nondem_meshdata.crs):
                    utils.reproject(nondem_meshdata, epsg4326)

            nondem_shape = utils.get_mesh_polygons(hfun.mesh.meshdata)
            if not epsg4326.equals(hfun.crs):
                transformer = Transformer.from_crs(
                    hfun.crs, epsg4326, always_xy=True)
                nondem_shape = ops.transform(
                        transformer.transform, nondem_shape)

            # In fast method all DEM hfuns have more priority than all
            # other inputs
            if big_cut_shape:
                nondem_meshdata = utils.clip_mesh_by_shape(
                    nondem_meshdata,
                    big_cut_shape,
                    use_box_only=False,
                    fit_inside=True,
                    inverse=True)

            for ishp in nondem_shape_list:
                nondem_meshdata = utils.clip_mesh_by_shape(
                    nondem_meshdata,
                    ishp,
                    use_box_only=False,
                    fit_inside=True,
                    inverse=True)

            nondem_shape_list.append(nondem_shape)

            if nondem_meshdata.tria is not None:
                index.append(nondem_meshdata.tria + offset)
            coord.append(nondem_meshdata.coords)
            value.append(nondem_meshdata.values)
            offset += coord[-1].shape[0]

        # Stack arrays
        final_coords = np.vstack(coord) if coord else np.empty((0, 2))
        final_values = np.vstack(value) if value else np.empty((0, 1))
        # Handle case where some inputs might not have triangles (points only)
        if index:
            final_tria = np.vstack(index)
        else:
            final_tria = None

        composite_hfun = MeshData(
            coords=final_coords,
            tria=final_tria,
            values=final_values,
            crs=epsg4326
        )

        hmin = self._size_info['hmin']
        hmax = self._size_info['hmax']
        if hmin:
            composite_hfun.values[composite_hfun.values < hmin] = hmin
        if hmax:
            composite_hfun.values[composite_hfun.values > hmax] = hmax

        utils.project_to_utm(composite_hfun)

        return composite_hfun



from typing import List, Optional, Union

import numpy as np
from pyproj import CRS


class MeshData:
    """
    A simple container for mesh data to replace jigsawpy_msh_t in OCSMesh.

    Parameters
    ----------
    coords : array_like
        Nodal coordinates with shape (K, 2).
    tria : array_like, optional
        Connectivity array for triangles with shape (N, 3). If None, an empty
        array is initialized.
    quad : array_like, optional
        Connectivity array for quads with shape (M, 4). If None, an empty
        array is initialized.
    values : array_like, optional
        Nodal values/attributes with shape (K,) for scalars or (K, D) for
        vectors. If None, initialized to (K,) zeros (scalar default).
    crs : Union[str, CRS], optional
        Coordinate Reference System. Can be a string (e.g., "EPSG:4326") or
        a pyproj.CRS object. Defaults to None.

    Attributes
    ----------
    coords : ndarray
        Nodal coordinates with shape (K, 2).
    tria : ndarray
        Triangle connectivity array with shape (N, 3).
    quad : ndarray
        Quad connectivity array with shape (M, 4).
    values : ndarray
        Nodal values/attributes with shape (K,) or (K, D).
    crs : CRS
        The coordinate reference system of the mesh.
    num_nodes : int
        Number of nodes in the mesh.
    """

    def __init__(
        self,
        coords: Union[np.ndarray, List],
        tria: Optional[Union[np.ndarray, List]] = None,
        quad: Optional[Union[np.ndarray, List]] = None,
        values: Optional[Union[np.ndarray, List]] = None,
        crs: Optional[Union[str, CRS]] = None
    ):
        # Initialize internal storage variables
        self._coords = None
        self._tria = None
        self._quad = None
        self._values = None
        self._crs = None

        # Set Coordinates (Required)
        # Uses the setter defined below to handle validation immediately
        # This will also initialize self._values to zeros of the correct size
        self.coords = coords

        # Initialize Triangles (Optional, defaults to empty)
        if tria is not None:
            self.tria = tria
        else:
            self._tria = np.empty((0, 3), dtype=int)

        # Initialize Quads (Optional, defaults to empty)
        if quad is not None:
            self.quad = quad
        else:
            self._quad = np.empty((0, 4), dtype=int)

        # Initialize Values (Optional)
        # If provided, overwrite the default zeros created by coords setter
        if values is not None:
            self.values = values

        # Initialize CRS (Optional)
        self.crs = crs

    def __repr__(self):
        # Improve repr to show vector dimension if applicable
        val_shape = self.values.shape if self.values is not None else (0,)
        crs_info = self.crs.to_string() if self.crs else "None"
        return (
            f"<MeshData: "
            f"{self.num_nodes} Nodes, "
            f"{self.tria.shape[0]} Triangles, "
            f"{self.quad.shape[0]} Quads, "
            f"Values Shape: {val_shape}, "
            f"CRS: {crs_info}>"
        )

    @property
    def num_nodes(self) -> int:
        """
        The number of nodes in the mesh.

        Returns
        -------
        int
            The count of rows in the coordinates array.
        """
        if self._coords is not None:
            return self._coords.shape[0]
        return 0

    # ==========================
    # Coordinates Property
    # ==========================
    @property
    def coords(self) -> np.ndarray:
        """
        Nodal coordinates array.

        Returns
        -------
        ndarray
            Array of shape (K, 2).

        Raises
        ------
        ValueError
            If assigned array is not 2-dimensional or does not have exactly
            2 columns.
        """
        return self._coords

    @coords.setter
    def coords(self, new_coords: Union[np.ndarray, List]):
        arr = np.array(new_coords, dtype=float)

        # Validation: Must be 2D array
        if arr.ndim != 2:
            raise ValueError("Coords array must be 2-dimensional.")

        # Validation: Must have exactly 2 columns (x, y)
        if arr.shape[1] != 2:
            raise ValueError(
                f"Coords must have 2 columns, got {arr.shape[1]}."
            )

        self._coords = arr

        # Enforce consistency: Ensure values array matches new node count
        if self._values is None or self._values.shape[0] != self.num_nodes:
            # If values exist preserve width
            width = 1
            if self._values is not None:
                width = self._values.shape[1]
            self._values = np.zeros((self.num_nodes, width), dtype=float)

    # ==========================
    # Triangles Property
    # ==========================
    @property
    def tria(self) -> np.ndarray:
        """
        Triangle connectivity array.

        Returns
        -------
        ndarray
            Array of shape (N, 3).

        Raises
        ------
        ValueError
            If assigned array does not have 3 columns or contains non-integers.
        """
        return self._tria

    @tria.setter
    def tria(self, new_tria: Union[np.ndarray, List]):
        # Create array without forcing dtype=int yet to check for decimals
        arr = np.array(new_tria)

        # Allow empty assignment
        if arr.size == 0:
            self._tria = np.empty((0, 3), dtype=int)
            return

        # Check for non-integer values (e.g. 1.5)
        if np.issubdtype(arr.dtype, np.floating) and not np.all(arr == arr.astype(int)):
            raise ValueError("Triangle elements must be integers.")
        # Safe to cast now
        arr = arr.astype(int)

        # Validation: Must have 3 columns
        if arr.ndim != 2 or arr.shape[1] != 3:
            raise ValueError(
                f"Triangles must be shape (N, 3), got shape {arr.shape}."
            )

        self._tria = arr

    # ==========================
    # Quads Property
    # ==========================
    @property
    def quad(self) -> np.ndarray:
        """
        Quad connectivity array.

        Returns
        -------
        ndarray
            Array of shape (M, 4).

        Raises
        ------
        ValueError
            If assigned array does not have 4 columns or contains non-integers.
        """
        return self._quad

    @quad.setter
    def quad(self, new_quad: Union[np.ndarray, List]):
        # Create array without forcing dtype=int yet to check for decimals
        arr = np.array(new_quad)

        # Allow empty assignment
        if arr.size == 0:
            self._quad = np.empty((0, 4), dtype=int)
            return

        # Check for non-integer values (e.g. 1.5)
        if np.issubdtype(arr.dtype, np.floating) and not np.all(arr == arr.astype(int)):
            raise ValueError("Quad elements must be integers.")
        # Safe to cast now
        arr = arr.astype(int)

        # Validation: Must have 4 columns
        if arr.ndim != 2 or arr.shape[1] != 4:
            raise ValueError(
                f"Quads must be shape (N, 4), got shape {arr.shape}."
            )

        self._quad = arr

    # ==========================
    # Values Property
    # ==========================
    @property
    def values(self) -> np.ndarray:
        """
        Nodal values array.

        Returns
        -------
        ndarray
            Array of shape (K,) for scalars or (K, D) for vectors.

        Raises
        ------
        ValueError
            If the length (first dimension) of the values array does not
            match `num_nodes`.
        """
        return self._values

    @values.setter
    def values(self, new_values: Union[np.ndarray, List, float]):
        arr = np.array(new_values, dtype=float)
        if arr.size == 1:
            arr = np.full(self.num_nodes, new_values, dtype=float)

        # Always a 2D array
        if arr.ndim == 1:
            arr = arr.reshape(-1, 1)
        elif arr.ndim > 2:
            raise ValueError(
                f"Values must be shape (N, M), got shape {arr.shape}."
            )

        # Validation: Check length against number of nodes
        if self._coords is not None and arr.shape[0] != self.num_nodes:
            raise ValueError(
                f"Values length {arr.shape[0]} does not match number of "
                f"nodes {self.num_nodes}."
            )

        self._values = arr

    # ==========================
    # CRS Property
    # ==========================
    @property
    def crs(self) -> Optional[CRS]:
        """
        Coordinate Reference System.

        Returns
        -------
        pyproj.CRS or None
        """
        return self._crs

    @crs.setter
    def crs(self, new_crs: Optional[Union[str, CRS]]):
        if new_crs is None:
            self._crs = None
        else:
            # Validate and convert using pyproj
            self._crs = CRS.from_user_input(new_crs)



import unittest

import numpy as np
from pyproj.crs import CRS, CRSError

from ocsmesh import MeshData


class TestMeshData(unittest.TestCase):

    def setUp(self):
        """Set up some basic data for testing."""
        self.coords_2d = np.array([
            [0.0, 0.0],
            [1.0, 0.0],
            [0.0, 1.0],
            [1.0, 1.0]
        ])
        self.tria = np.array([[0, 1, 2]])
        self.quad = np.array([[0, 1, 3, 2]])
        # Scalar values for basic setup
        self.values = np.array([10.0, 20.0, 30.0, 40.0])

    def test_initialization_full(self):
        """Test initializing with all arguments provided."""
        mesh = MeshData(
            coords=self.coords_2d,
            tria=self.tria,
            quad=self.quad,
            values=self.values
        )

        self.assertEqual(mesh.num_nodes, 4)
        np.testing.assert_array_equal(mesh.coords, self.coords_2d)
        np.testing.assert_array_equal(mesh.tria, self.tria)
        np.testing.assert_array_equal(mesh.quad, self.quad)
        np.testing.assert_array_equal(mesh.values, self.values[:, None])

    def test_initialization_minimal(self):
        """Test initializing with only coordinates (required)."""
        mesh = MeshData(coords=self.coords_2d)

        self.assertEqual(mesh.num_nodes, 4)
        # Should default to empty arrays with correct column counts
        self.assertEqual(mesh.tria.shape, (0, 3))
        self.assertEqual(mesh.quad.shape, (0, 4))
        # Should default to zeros (K, 1)
        self.assertEqual(mesh.values.shape, (4, 1))
        self.assertTrue(np.all(mesh.values == 0))

    def test_vector_values(self):
        """Test assigning vector values (K, D) where D > 1."""
        # 2D vector per node (e.g., velocity u, v)
        vector_vals = np.array([
            [1.0, 0.0],
            [0.0, 1.0],
            [-1.0, 0.0],
            [0.0, -1.0]
        ])
        mesh = MeshData(coords=self.coords_2d, values=vector_vals)
        self.assertEqual(mesh.values.shape, (4, 2))
        np.testing.assert_array_equal(mesh.values, vector_vals)

    def test_explicit_column_vector(self):
        """
        Test assigning (K, 1) values.  Ensure they remain (K, 1) 
        """
        col_vec = np.array([[1], [2], [3], [4]])
        mesh = MeshData(coords=self.coords_2d, values=col_vec)
        self.assertEqual(mesh.values.shape, (4, 1))

    def test_vector_preservation_on_resize(self):
        """
        Test that if we have vector values and resize the mesh,
        the new zero-values array maintains the vector dimension.
        """
        # Initialize with vectors (4 nodes, 3 components)
        vector_vals = np.zeros((4, 3))
        mesh = MeshData(coords=self.coords_2d, values=vector_vals)

        # Resize mesh to 2 nodes
        mesh.coords = [[0, 0], [1, 1]]

        self.assertEqual(mesh.num_nodes, 2)
        # Should be (2, 3) - width 3 preserved
        self.assertEqual(mesh.values.shape, (2, 3))
        self.assertTrue(np.all(mesh.values == 0))

    def test_store_scalar_as_column_vector(self):
        """
        Test that if we pass scalar values (K,) it is stored
        as (K, 1)
        """
        mesh = MeshData(coords=self.coords_2d)  # Default (4,)
        self.assertEqual(mesh.values.ndim, 2)

    def test_coords_3d_rejected(self):
        """Test that 3D coordinates (x, y, z) are NOT accepted."""
        coords_3d = np.array([
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 1.0]
        ])
        # Should raise ValueError now
        with self.assertRaisesRegex(ValueError, "Coords must have 2 columns"):
            MeshData(coords=coords_3d)

    def test_values_length_mismatch(self):
        """Test setting values with mismatching length raises ValueError."""
        mesh = MeshData(coords=self.coords_2d)  # 4 nodes

        # Try to set 3 vectors for 4 nodes
        vector_mismatch = np.zeros((3, 2))
        with self.assertRaises(ValueError):
            mesh.values = vector_mismatch

        # Try to set 5 scalars for 4 nodes
        with self.assertRaises(ValueError):
            mesh.values = [1, 2, 3, 4, 5]

    def test_coords_update_logic(self):
        """
        Test that updating coordinates handles values array correctly:
        1. Resizes (resets) values if node count changes.
        2. Preserves values if node count stays same.
        """
        mesh = MeshData(coords=self.coords_2d, values=[1, 2, 3, 4])

        # 1. Update coords with DIFFERENT size (2 nodes instead of 4)
        mesh.coords = [[0, 0], [1, 1]]
        self.assertEqual(mesh.num_nodes, 2)
        # Values should be reset to zeros column vector of length 2
        self.assertEqual(mesh.values.shape, (2, 1))
        self.assertTrue(np.all(mesh.values == 0))

        # 2. Update coords with SAME size (2 nodes)
        # Set specific values first
        mesh.values = [99, 100]
        # Update coords
        mesh.coords = [[5, 5], [6, 6]]
        # Values should remain [99, 100]
        np.testing.assert_array_equal(mesh.values, [[99], [100]])

    def test_element_integer_validation(self):
        """Test that elements are rejected if they contain decimals."""
        mesh = MeshData(coords=self.coords_2d)

        # 1. Triangles
        # Valid floats (e.g. 1.0) should pass and convert to int
        mesh.tria = [[0.0, 1.0, 2.0]]
        self.assertTrue(np.issubdtype(mesh.tria.dtype, np.integer))
        self.assertEqual(mesh.tria[0, 0], 0)

        # Invalid floats (e.g. 0.5) should raise ValueError
        with self.assertRaisesRegex(ValueError, "Triangle elements must be integers"):
            mesh.tria = [[0.5, 1.0, 2.0]]

        # 2. Quads
        # Valid floats
        mesh.quad = [[0.0, 1.0, 2.0, 3.0]]
        self.assertTrue(np.issubdtype(mesh.quad.dtype, np.integer))

        # Invalid floats
        with self.assertRaisesRegex(ValueError, "Quad elements must be integers"):
            mesh.quad = [[0.1, 1.0, 2.0, 3.0]]

    def test_list_input_compatibility(self):
        """Test that Python lists are correctly converted to numpy arrays."""
        mesh = MeshData(
            coords=[[0, 0], [1, 1]],
            tria=[[0, 1, 0]]
        )
        self.assertIsInstance(mesh.coords, np.ndarray)
        self.assertIsInstance(mesh.tria, np.ndarray)

    def test_crs_assignment(self):
        """Test valid and invalid CRS assignments."""
        mesh = MeshData(coords=self.coords_2d)

        # 1. Default None
        self.assertIsNone(mesh.crs)

        # 2. Assign Valid String
        mesh.crs = "EPSG:4326"
        self.assertIsInstance(mesh.crs, CRS)
        self.assertEqual(mesh.crs.to_string(), "EPSG:4326")

        # 3. Assign Valid CRS Object
        proj_crs = CRS.from_epsg(32610) # UTM Zone 10N
        mesh.crs = proj_crs
        self.assertIsInstance(mesh.crs, CRS)
        self.assertEqual(mesh.crs, proj_crs)

        # 4. Assign None
        mesh.crs = None
        self.assertIsNone(mesh.crs)

        # 5. Assign Invalid
        with self.assertRaises(CRSError):
            mesh.crs = "INVALID:12345"
            
    def test_crs_init(self):
        """Test CRS initialization via constructor."""
        mesh = MeshData(coords=self.coords_2d, crs="EPSG:3857")
        self.assertIsInstance(mesh.crs, CRS)
        self.assertEqual(mesh.crs.to_string(), "EPSG:3857")


if __name__ == '__main__':
    unittest.main()



import os
import tempfile
import urllib.request
from pathlib import Path

from rasterio.enums import Resampling

from ocsmesh.raster import Raster


# Find a better way!
tif_url = (
    'https://chs.coast.noaa.gov/htdata/raster2/elevation/NCEI_ninth_Topobathy_2014_8483/northeast_sandy/ncei19_n40x75_w073x75_2015v1.tif'
)
TEST_FILE = os.path.join(tempfile.gettempdir(), 'test_dem.tif')
if not Path(TEST_FILE).exists():
    tmpfd, tmppath = tempfile.mkstemp()
    urllib.request.urlretrieve(tif_url, filename=tmppath)
    os.close(tmpfd)
    r = Raster(tmppath)
    r.resampling_method = Resampling.average
    r.resample(scaling_factor=0.01)
    r.save(TEST_FILE)



#!/bin/env python3
from copy import deepcopy
import logging
from pathlib import Path
import sys
from time import time

import geopandas as gpd
import numpy as np
from pyproj import CRS, Transformer
from scipy.interpolate import griddata
from scipy.spatial import cKDTree
from shapely.geometry import MultiPolygon, Polygon, GeometryCollection
from shapely.ops import polygonize, unary_union, transform

from ocsmesh import Raster, Geom, Mesh, Hfun, utils
from ocsmesh.internal import MeshData
from ocsmesh.engines.factory import get_mesh_engine

logging.basicConfig(
    stream=sys.stdout,
    format='%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%Y-%m-%d:%H:%M:%S'
    )
#logging.getLogger().setLevel(logging.DEBUG)
#logging.getLogger().setLevel(logging.INFO)

_logger = logging.getLogger(__name__)

class SubsetAndCombine:

    @property
    def script_name(self):
        return 'subset_n_combine'

    def __init__(self, sub_parser):

        this_parser = sub_parser.add_parser(
            self.script_name,
            help="Subset and combine the two input meshes based on the region of interset")

        this_parser.add_argument(
                "--rasters", nargs='+', type=Path,
                help="List of rasters used for calculating depth cutoff")
        this_parser.add_argument(
                "-o", "--out", type=Path, metavar="DIRECTORY",
                help="Output directory")
        this_parser.add_argument(
                "-a", "--outall", action='store_true',
                help="Flag to indicate all intermediate shapes to be written to disk")
        this_parser.add_argument(
                "-c", "--crs", default="epsg:4326",
                help="Input mesh CRS and output CRS [experimental]")
        this_parser.add_argument(
                "-d", "--cutoff", type=float, default=-200,
                help="Cut-off depth (positive up) for calculating"
                     + " region subsetted from fine mesh")
        this_parser.add_argument(
                "-s", "--isotach-speed", metavar="WIND_SPEED",
                choices=[34, 50, 64], type=int, default=34,
                help="Isotach (34, 50 or 64) to use if region of interset is storm track")
        this_parser.add_argument(
                "-n", "--adv-buffer-n-layers", metavar="N_LAYERS",
                type=int, default=9,
                help="Number of buffer layers to mesh between fine and coarse mesh")
        this_parser.add_argument(
                "--adv-upstream-area-max", metavar="AREA",
                type=float, default=100,
                help="Advanced option: maximum segment area for calculating upstream")
        this_parser.add_argument(
                "--adv-rel-island-area-min", metavar="AREA",
                type=float, default=1/100,
                help="Advanced option: minimum segment area for"
                     + " dropping islands during shape cleanup")
        this_parser.add_argument(
                "fine_mesh", type=Path,
                help="Path to the fine mesh")
        this_parser.add_argument(
                "coarse_mesh", type=Path,
                help="Path to the coarse mesh")
        this_parser.add_argument(
                "region_of_interset", type=Path,
                help="Path to the shape file for region of interest or storm track")

    def run(self, args):

        # Currently only used for clipping!
        pathlist_raster = args.rasters

        path_fine = args.fine_mesh
        path_coarse = args.coarse_mesh

        # e.g. track - test it's shapefile
        roi = args.region_of_interset

        out_dir = args.out

        # TODO: Clarify what "crs" refers to (it's both input and output!)
        crs = CRS.from_user_input(args.crs)

        wind_speed = args.isotach_speed
        cutoff_elev = args.cutoff

        # 9 layers: fine mesh coarser than coarse mesh in some areas!
        num_buffer_layers = args.adv_buffer_n_layers

        upstream_size_max = args.adv_upstream_area_max
        rel_island_area_min = args.adv_rel_island_area_min

        out_all = args.outall

        self._main(
            pathlist_raster, path_fine, path_coarse, roi,
            cutoff_elev, upstream_size_max,
            wind_speed, num_buffer_layers, rel_island_area_min,
            out_dir, crs, out_all)

    def _get_largest_polygon(self, mpoly):
        if isinstance(mpoly, Polygon):
            return mpoly

        return sorted(mpoly.geoms, key=lambda i: i.area)[-1]

    def _get_polygons_smaller_than(self, mpoly, size):
        if isinstance(mpoly, Polygon):
            if mpoly.area <= size:
                return mpoly
            return Polygon()

        areas = np.array([p.area for p in mpoly.geoms])

        return MultiPolygon(np.extract(areas < size, mpoly.geoms).tolist())



    def _get_region_of_interest(self, track_file, wind_speed=34):

        # CRS is assumed to be epsg 4326
        crs = 4326
        gdf_roi = gpd.read_file(track_file)
        if not (gdf_roi.crs is None or gdf_roi.crs.equals(crs)):
            gdf_roi = gdf_roi.to_crs(crs)
        if "RADII" in gdf_roi.columns:
            gdf_specific_isotach = gdf_roi[
                    gdf_roi.RADII.astype(int) == wind_speed]
            isotach_exterior_polygon = MultiPolygon(
                list(polygonize(list(gdf_specific_isotach.exterior))))
            region_of_interest = isotach_exterior_polygon

        else:
            region_of_interest = gdf_roi.union_all()


        return region_of_interest

    def _calculate_clipping_polygon(
            self,
            pathlist_raster,
            region_of_interest, crs,
            cutoff_elev=-200,
            upstream_size_max=100,
            upstream_poly_list=None):

        rasters = [Raster(raster_file) for raster_file in pathlist_raster]
        geom_cutoff = Geom(
                rasters, zmin=cutoff_elev,
                base_shape=region_of_interest,
                base_shape_crs=crs)
        clip_poly_draft = geom_cutoff.get_multipolygon()

        t = Transformer.from_crs(4326, crs, always_xy=True)
        clip_poly_draft = transform(t.transform, clip_poly_draft)

        # Add back upstream
        poly_upstreams = []
        if upstream_poly_list is not None:
            for poly in upstream_poly_list:
                poly_upstreams.append(
                    self._get_polygons_smaller_than(
                        poly.difference(clip_poly_draft),
                        upstream_size_max))

        # Islands are not of intereset when clipping high and low-res meshes
        clip_poly = utils.remove_holes(unary_union([
            clip_poly_draft, *poly_upstreams]))

        return clip_poly



    def _get_polygon_from_geom_collection(self, shape):

        if not isinstance(shape, GeometryCollection):
            raise ValueError(
                f"Expected a GeometryCollection, received {type(shape)}")

        polygon_list = []
        for g in shape.geoms:
            if isinstance(g, Polygon):
                polygon_list.append(g)
            elif isinstance(g, MultiPolygon):
                polygon_list.extend(g.geoms)
        if len(polygon_list) == 0:
            raise ValueError("No polygon in the intersection!")

        return MultiPolygon(polygon_list)

    def _add_one_mesh_layer_to_polygon(
            self,
            polygon,
            fine_mesh, fine_polygon,
            coarse_mesh, coarse_polygon):

        fine_clip_plus_layer = utils.clip_mesh_by_shape(
                fine_mesh, fine_polygon,
                fit_inside=False, adjacent_layers=1)
        poly_1_lyr_hires = utils.get_mesh_polygons(
                fine_clip_plus_layer).difference(fine_polygon)

        coarse_clip_plus_layer = utils.clip_mesh_by_shape(
                coarse_mesh, coarse_polygon,
                fit_inside=False, adjacent_layers=1)
        poly_1_lyr_lowres = utils.get_mesh_polygons(
                coarse_clip_plus_layer).difference(coarse_polygon)

        # Add the one layer to the buffer to make boundaries shape conformal
        return unary_union(
                [polygon, poly_1_lyr_hires, poly_1_lyr_lowres])


    def _add_overlap_to_polygon(self, mesh, polygon):

        # Find segment to add to polygon
        clipped_add = utils.clip_mesh_by_shape(
                mesh, polygon,
                fit_inside=False,
                inverse=False,
                check_cross_edges=True)
        poly_add = utils.get_mesh_polygons(clipped_add)
        new_polygon = unary_union([polygon, poly_add])

        # Inverse of above
        clipped_mesh = utils.clip_mesh_by_shape(
                mesh, polygon,
                fit_inside=False,
                inverse=True,
                check_cross_edges=True
                )

        return new_polygon, clipped_mesh

    def _calculate_mesh_size_function(
            self,
            buffer_domain,
            hires_mesh_clip,
            lores_mesh_clip,
            buffer_crs
        ):

        assert buffer_crs == hires_mesh_clip.crs == lores_mesh_clip.crs

        # HARDCODED FOR NOW
        approx_elem_per_width = 3

        meshdata_hi = deepcopy(hires_mesh_clip)
        meshdata_lo = deepcopy(lores_mesh_clip)

        assert not buffer_crs.is_geographic

        # calculate mesh size for clipped bits
        hfun_hi = Hfun(Mesh(meshdata_hi))
        hfun_hi.size_from_mesh()

        hfun_lo = Hfun(Mesh(meshdata_lo))
        hfun_lo.size_from_mesh()

        engine = get_mesh_engine('triangle', opts='p')
        meshdata_cdt = engine.generate(gpd.GeoSeries(buffer_domain))
        meshdata_cdt.crs = buffer_crs

        hfun_cdt = Hfun(Mesh(meshdata_cdt))
        hfun_cdt.size_from_mesh()

        hfun_cdt_sz = deepcopy(hfun_cdt.meshdata().values) / approx_elem_per_width
        meshdata_cdt.values = hfun_cdt_sz

        # TODO: Make jigsaw an option
        engine = get_mesh_engine('jigsaw')
        # Note we alraeady know the buffer_crs is NOT geographic
        meshdata_domain_rep = engine.generate(
            gpd.GeoSeries(buffer_domain),
            meshdata_cdt,
        )
        utils.finalize_mesh(meshdata_domain_rep)


        pts_2mesh = np.vstack(
            (hfun_hi.meshdata().coords, hfun_lo.meshdata().coords)
        )
        val_2mesh = np.vstack(
            (hfun_hi.meshdata().values, hfun_lo.meshdata().values)
        )
        domain_sz_1 = griddata(
            pts_2mesh, val_2mesh, meshdata_domain_rep.coords, method='linear'
        )
        domain_sz_2 = griddata(
            pts_2mesh, val_2mesh, meshdata_domain_rep.coords, method='nearest'
        )
        domain_sz = domain_sz_1.copy()
        domain_sz[np.isnan(domain_sz_1)] = domain_sz_2[np.isnan(domain_sz_1)]

        meshdata_domain_rep.values = domain_sz
        meshdata_domain_rep.crs = buffer_crs

        return meshdata_domain_rep


    def _generate_mesh_for_buffer_region(
            self, buffer_polygon, meshdata_hfun_buffer, buffer_crs):

        assert not buffer_crs.is_geographic
        assert buffer_crs == meshdata_hfun_buffer.crs

        # TODO: Make jigsaw an option
        engine = get_mesh_engine('jigsaw')
        # Note we alraeady know the buffer_crs is NOT geographic
        meshdata_buf_apprx = engine.generate(
            gpd.GeoSeries(buffer_polygon),
            meshdata_hfun_buffer,
        )
        utils.finalize_mesh(meshdata_buf_apprx)

        # If vertices are too close to buffer geom boundary,
        # it's going to cause issues (thin elements)
#        if meshdata_buf_apprx.crs != hfun_buffer.crs:
#            utils.reproject(meshdata_buf_apprx, hfun_buffer.crs)
        seed_mesh = utils.clip_mesh_by_shape(
            meshdata_buf_apprx,
            buffer_polygon.buffer(-min(meshdata_hfun_buffer.values).item())
        )

#        utils.reproject(meshdata_buf_apprx, buffer_crs)
        engine = get_mesh_engine('triangle', opts='p')
        meshdata_buffer = engine.generate(
            gpd.GeoSeries(buffer_polygon),
            seed=seed_mesh
        )
        meshdata_buffer.crs = buffer_crs

#        utils.reproject(meshdata_buffer, buffer_crs)

        return meshdata_buffer


    def _transform_mesh(self, mesh, out_crs):

        transformer = Transformer.from_crs(
            mesh.crs, out_crs, always_xy=True)

        coords = mesh.coords

        # pylint: disable=E0633
        coords[:, 0], coords[:, 1] = transformer.transform(
                coords[:, 0], coords[:, 1])
        mesh.coords[:] = coords
        mesh.crs = out_crs


    def _merge_all_meshes(self, crs, meshdata_mesh, meshdata_clip_lowres, meshdata_clip_hires):

        # Combine mesh
        # 1. combine hires and low-res
        meshdata_old = utils.merge_meshdata(
            meshdata_clip_lowres, meshdata_clip_hires,
            drop_by_bbox=False, out_crs=crs)

        # 2. Create kdtree for boundary of combined mesh
        old_mesh_bdry_edges = utils.get_boundary_edges(meshdata_old)
        old_mesh_bdry_verts = np.unique(old_mesh_bdry_edges)
        old_mesh_bdry_coords = meshdata_old.coords[old_mesh_bdry_verts]

        tree_old = cKDTree(old_mesh_bdry_coords)

        # 3. Find shared boundary nodes from the tree and bdry nodes
        new_mesh_bdry_edges = utils.get_boundary_edges(meshdata_mesh)
        new_mesh_bdry_verts = np.unique(new_mesh_bdry_edges)
        new_mesh_bdry_coords = meshdata_mesh.coords[new_mesh_bdry_verts]

        tree_new = cKDTree(new_mesh_bdry_coords)

        neigh_idxs = tree_old.query_ball_tree(tree_new, r=1e-6)

        # 4. Create a map for shared nodes
        map_idx_shared = {}
        for idx_tree_old, neigh_idx_list in enumerate(neigh_idxs):
            num_match = len(neigh_idx_list)
            if num_match == 0:
                continue
            if num_match > 1:
                raise ValueError("More than one node match on boundary!")

            idx_tree_new = neigh_idx_list[0]

            map_idx_shared[new_mesh_bdry_verts[idx_tree_new]] = old_mesh_bdry_verts[idx_tree_old]

        # 5. Combine seam into the rest replacing the index for shared nodes
        #    with the ones from tree
        mesh_types = {
            'tria': int,
            'quad': int
        }

        coord = []
        elems = {k: [] for k in mesh_types}
        value = []
        offset = 0

        for k in mesh_types:
            elems[k].append(getattr(meshdata_old, k) + offset)
        coord.append(meshdata_old.coords)
        value.append(meshdata_old.values)
        offset += coord[-1].shape[0]

        # Drop shared vertices and update element cnn based on map and dropped offset
        mesh_orig_idx = np.arange(len(meshdata_mesh.coords))
        mesh_shrd_idx = np.unique(list(map_idx_shared.keys()))
        mesh_renum_idx = np.setdiff1d(mesh_orig_idx, mesh_shrd_idx)
        map_to_combined_idx = {
            index: i + offset for i, index in enumerate(mesh_renum_idx)}
        map_to_combined_idx.update(map_idx_shared)

        for k in mesh_types:
            cnn = getattr(meshdata_mesh, k)
            # If it's empty executing list comprehension results in a
            # (0,) shape instead of (0, 4)
            if cnn.shape[0] == 0:
                elems[k].append(cnn)
                continue

            elems[k].append(np.array([[map_to_combined_idx[x]
                                      for x in  elm] for elm in cnn]))

        coord.append(meshdata_mesh.coords[mesh_renum_idx, :])
        value.append(meshdata_mesh.values[mesh_renum_idx])

        # Putting it all together
        composite_mesh = MeshData(
            coords=np.vstack(coord),
            values=np.vstack(value),
            crs=crs,
            **{k: np.vstack(v) for k, v in elems.items()},
        )

        return composite_mesh, mesh_shrd_idx


    def _write_polygon(self, polygon, crs, path):
        gdf  = gpd.GeoDataFrame(geometry=gpd.GeoSeries(polygon), crs=crs)
        gdf.to_file(path)

    def _write_meshdata(self, meshdata_mesh, path):
        mesh_obj = Mesh(meshdata_mesh)
        mesh_obj.write(path, format="2dm", overwrite=True)

    def _write_outputs(
            self,
            out_dir,
            out_all,
            crs,
            poly_clip_hires,
            poly_clip_lowres,
            poly_seam,
            poly_clipper,
            mesh_fine,
            mesh_coarse,
            meshdata_buffer_mesh,
            meshdata_clip_hires,
            meshdata_clip_lowres,
            meshdata_combined_mesh,
            buffer_shrd_idx
            ):

        if out_all:
            _logger.info("Writing shapes...")
            start = time()
            self._write_polygon(poly_clip_hires, crs, out_dir/"hires")
            self._write_polygon(poly_clip_lowres, crs, out_dir/"lowres")
            self._write_polygon(poly_seam, crs, out_dir/"seam")
            self._write_polygon(poly_clipper, crs, out_dir/"clip")
            _logger.info(f"Done in {time() - start} sec")

            _logger.info("Writing input mesh (copy)...")
            start = time()
            mesh_fine.write(str(out_dir / "fine.2dm"), format="2dm", overwrite=True)
            mesh_coarse.write(str(out_dir / "coarse.2dm"), format="2dm", overwrite=True)
            _logger.info(f"Done in {time() - start} sec")

            _logger.info("Writing mesh...")
            start = time()
            self._write_meshdata(meshdata_buffer_mesh, out_dir/"seam_mesh.2dm")
            self._write_meshdata(meshdata_clip_hires, out_dir/"hires_mesh.2dm")
            self._write_meshdata(meshdata_clip_lowres, out_dir/"lowres_mesh.2dm")
            _logger.info(f"Done in {time() - start} sec")

        _logger.info("Writing mesh...")
        start = time()
        self._write_meshdata(meshdata_combined_mesh, out_dir/"final_mesh.2dm")
        _logger.info(f"Done in {time() - start} sec")


    def _interpolate_values(self, meshdata_combined_mesh, mesh_fine, mesh_coarse):
        interp_1_meshdata = deepcopy(meshdata_combined_mesh)
        utils.interpolate_euclidean_mesh_to_euclidean_mesh(
            mesh_fine.meshdata, interp_1_meshdata,
            method='linear', fill_value=np.nan)

        interp_2_meshdata = deepcopy(meshdata_combined_mesh)
        utils.interpolate_euclidean_mesh_to_euclidean_mesh(
            mesh_coarse.meshdata, interp_2_meshdata,
            method='linear', fill_value=np.nan)


        # Manually get the nan values and only overwrite them!
        mask_1 = np.isnan(interp_1_meshdata.values)
        mask_2 = np.logical_not(np.isnan(interp_2_meshdata.values))
        mask = np.logical_and(mask_1, mask_2)

        meshdata_combined_mesh.values = interp_1_meshdata.values
        meshdata_combined_mesh.values[mask] = interp_2_meshdata.values[mask]


    def _main(
            self,
            pathlist_raster, path_fine, path_coarse, track_file,
            cutoff_elev, upstream_size_max,
            wind_speed, num_buffer_layers, rel_island_area_min,
            out_dir, crs, out_all):

        _logger.info("Reading meshes...")
        start = time()
        mesh_fine = Mesh.open(path_fine, crs=crs)
        mesh_coarse = Mesh.open(path_coarse, crs=crs)
        _logger.info(f"Done in {time() - start} sec")



        _logger.info("Calculate impact area...")
        start = time()

        # poly_isotach is in EPSG:4326
        poly_isotach = self._get_region_of_interest(
            track_file, wind_speed
        )
        utm = utils.estimate_bounds_utm(poly_isotach.bounds, 4326)

        # Transform all inputs to UTM:
        t1 = Transformer.from_crs(4326, utm, always_xy=True)
        poly_isotach = transform(t1.transform, poly_isotach)
        utils.reproject(mesh_fine.meshdata, utm)
        utils.reproject(mesh_coarse.meshdata, utm)


        _logger.info("Calculate mesh polygons...")
        start = time()
        poly_fine = self._get_largest_polygon(mesh_fine.get_multipolygon())
        poly_coarse = self._get_largest_polygon(mesh_coarse.get_multipolygon())
        _logger.info(f"Done in {time() - start} sec")

        poly_storm_roi = poly_isotach.intersection(poly_fine)

        poly_clipper = self._calculate_clipping_polygon(
            pathlist_raster=pathlist_raster,
            region_of_interest=poly_storm_roi,
            crs=utm,
            cutoff_elev=cutoff_elev,
            upstream_size_max=upstream_size_max,
            upstream_poly_list=[poly_fine])
        _logger.info(f"Done in {time() - start} sec")


        _logger.info("Calculate clipped polygons...")
        meshdata_clip_hires_0 = utils.clip_mesh_by_shape(
                mesh_fine.meshdata, poly_clipper,
                fit_inside=False)
        meshdata_clip_lowres_0 = utils.clip_mesh_by_shape(
                mesh_coarse.meshdata, poly_clipper,
                fit_inside=True,
                inverse=True,
                adjacent_layers=num_buffer_layers)
        _logger.info(f"Done in {time() - start} sec")

        _logger.info("Calculate clipped polygons...")
        start = time()
        poly_clip_hires_0 = utils.remove_holes(
                utils.get_mesh_polygons(meshdata_clip_hires_0))
        poly_clip_lowres_0 = utils.get_mesh_polygons(meshdata_clip_lowres_0)
        _logger.info(f"Done in {time() - start} sec")

        _logger.info("Calculating buffer region...")
        start = time()

        # Remove the two mesh clip regions from the coarse mesh polygon
        poly_seam_0 = poly_coarse.difference(
                unary_union([poly_clip_lowres_0, poly_clip_hires_0]))

        poly_seam_1 = poly_seam_0.intersection(poly_fine)

        # Get rid of non polygon results of the intersection
        poly_seam_2 = poly_seam_1
        if isinstance(poly_seam_1, GeometryCollection):
            poly_seam_2 = self._get_polygon_from_geom_collection(poly_seam_1)

        # Get one layer on each mesh
        poly_seam_3 = self._add_one_mesh_layer_to_polygon(
                poly_seam_2,
                mesh_fine.meshdata, poly_clip_hires_0,
                mesh_coarse.meshdata, poly_clip_lowres_0)


        # Attach overlaps to buffer region (due to 1 layer and upstream)
        poly_seam_4, meshdata_clip_hires_1 = self._add_overlap_to_polygon(
                meshdata_clip_hires_0, poly_seam_3)
        poly_seam_5, meshdata_clip_lowres_1 = self._add_overlap_to_polygon(
                meshdata_clip_lowres_0, poly_seam_4)

        # Cleanup buffer shape
        poly_seam_6 = utils.remove_holes_by_relative_size(
                poly_seam_5, rel_island_area_min)

        poly_seam_7 = utils.drop_extra_vertex_from_polygon(poly_seam_6)
        _logger.info(f"Done in {time() - start} sec")


        _logger.info("Calculate reclipped polygons...")
        start = time()
        meshdata_clip_hires = meshdata_clip_hires_1
        poly_clip_hires = utils.remove_holes(
                utils.get_mesh_polygons(meshdata_clip_hires))

        meshdata_clip_lowres = utils.clip_mesh_by_shape(
            meshdata_clip_lowres_1, poly_clip_hires,
            fit_inside=False, inverse=True)
        poly_clip_lowres = utils.get_mesh_polygons(meshdata_clip_lowres)
        _logger.info(f"Done in {time() - start} sec")

        poly_seam_8 = poly_seam_7.difference(
                    unary_union([poly_clip_hires, poly_clip_lowres]))


        poly_seam = poly_seam_8

        meshdata_hfun_buffer = self._calculate_mesh_size_function(
            poly_seam, meshdata_clip_hires, meshdata_clip_lowres, utm
        )
        meshdata_buffer_mesh = self._generate_mesh_for_buffer_region(
            poly_seam, meshdata_hfun_buffer, utm
        )

        _logger.info("Combining meshes...")
        start = time()
        meshdata_combined_mesh, buffer_shrd_idx = self._merge_all_meshes(
                utm, meshdata_buffer_mesh, meshdata_clip_lowres, meshdata_clip_hires)
        _logger.info(f"Done in {time() - start} sec")

        # NOTE: This call also detects overlap issues
        utils.finalize_mesh(meshdata_combined_mesh)

        self._interpolate_values(meshdata_combined_mesh, mesh_fine, mesh_coarse)

        # TODO: Interpolate DEM?

        self._write_outputs(
            out_dir,
            out_all,
            utm,
            poly_clip_hires,
            poly_clip_lowres,
            poly_seam,
            poly_clipper,
            mesh_fine,
            mesh_coarse,
            meshdata_buffer_mesh,
            meshdata_clip_hires,
            meshdata_clip_lowres,
            meshdata_combined_mesh,
            buffer_shrd_idx
        )



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Optional

import geopandas as gpd

from ocsmesh.internal import MeshData


class BaseMeshOptions(ABC):
    """
    Abstract base class for mesh engine options.
    """

    @abstractmethod
    def get_config(self) -> Any:
        """
        Returns the configuration object or structure
        required by the specific backend engine.
        """


class BaseMeshEngine(ABC):
    """
    Abstract base class for a mesh generation engine.
    """

    def __init__(self, options: BaseMeshOptions):
        """
        Initialize the engine with specific options.
        """
        if not isinstance(options, BaseMeshOptions):
            raise ValueError(
                f"options must be of type BaseMeshOptions, received {type(options)}"
            )
        self._options = options

    @abstractmethod
    def generate(
        self,
        shape: gpd.GeoSeries,
        sizing: Optional[MeshData | int | float] = None,
        seed: Optional[MeshData] = None,
    ) -> MeshData:
        """
        Generate a new mesh from a geometric shape.

        Parameters
        ----------
        shape : Any
            The input geometry (e.g., Polygon, MultiPolygon).
        sizing : Any, optional
            The sizing function or field.

        Returns
        -------
        MeshData
            The generated mesh.
        """

    @abstractmethod
    def remesh(
        self,
        mesh: MeshData,
        remesh_region: Optional[gpd.GeoSeries] = None,
        sizing: Optional[MeshData | int | float] = None,
        seed: Optional[MeshData] = None,
    ) -> MeshData:
        """
        Refine or optimize an existing mesh.

        Parameters
        ----------
        mesh : MeshData
            The input mesh to be modified.
        remesh_region : Any, optional
            The region to remesh. If None, remesh entire mesh.
        sizing : Any, optional
            The sizing function or field.

        Returns
        -------
        MeshData
            The resulting mesh.
        """
